/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB parameters and setup
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   Copyright (C) Paul Ashton 1997
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef _NT_DOMAIN_H /* _NT_DOMAIN_H */
#define _NT_DOMAIN_H 

/*
 * Okay, I don't like this being here, but the following includes
 * might need it
 */
typedef void PRS_HANDLE;


/* basic includes */
#include "util_list.h"

/* dce/rpc support */
#include "rpc_dce.h"

/* miscellaneous structures / defines */
#include "rpc_misc.h"

/* MS AD prototypes */
#include "sam.h"

/* 
 * A bunch of stuff that was put into smb.h
 * in the NTDOM branch - it didn't belong there.
 */

#define CHECK_STRUCT(data) \
{ \
	if ((data)->struct_start != 0xfefefefe || \
	    (data)->struct_end != 0xdcdcdcdc) \
	{ \
		DEBUG(0,("uninitialised structure (%s, %d)\n", \
		FUNCTION_MACRO, __LINE__)); \
		sleep(30); \
	} \
}

struct parse_struct;

typedef BOOL (* prs_align_fn) (struct parse_struct *ps, int object_size);

typedef struct parse_struct
{
	uint32 struct_start;

	char *data;		/* memory buffer */
	size_t data_size;	/* current memory buffer size */
	/* array memory offsets */
	uint32 start;
	uint32 end;

	uint32 offset;		/* offset currently being accessed in memory buffer */
	uint8 align;		/* data alignment */
	prs_align_fn align_fn;  /* function used for alignment of objects */
	BOOL io;		/* parsing in or out of data stream */
	BOOL error;		/* error occurred while parsing (out of memory bounds) */
	BOOL bigendian;		/* big-endian data */
	BOOL side;		/* client (False) or server (True) */

	char *last_level0;

	int depth; /* The current struct depth */
	GENERIC_LIST *pending_stack; /* list of pendings stack */

	uint32 parse_error;

	uint32 struct_end;

}
prs_struct;

/*
 * Defines for io member of prs_struct.
 */

#define MARSHALL 0
#define UNMARSHALL 1

#define MARSHALLING(ps) (!(ps)->io)
#define UNMARSHALLING(ps) ((ps)->io)

#define PRS_CLIENT 0
#define PRS_SERVER 1
#define PRS_IS_CLIENT(ps) (!(ps)->side)
#define PRS_IS_SERVER(ps) ((ps)->side)

typedef BOOL (* GenericParseCB) (const char *name, void *item, prs_struct *ps);


struct srv_auth_fns;

typedef struct rpcsrv_struct
{
	prs_struct data_i;	/* input data (intermediate, for fragments) */
	prs_struct rdata;	/* output data (to create fragments from */

	/* indicates how far in rdata we have got, creating fragments */
	uint32 rdata_offset;

	prs_struct smb_pdu;
	prs_struct rsmb_pdu;

	void *auth_info;
	struct srv_auth_fns *auth;

	/* set of authentication modules.  does not include noauth */
	uint32 num_auths;
	struct srv_auth_fns **auth_fns;

	BOOL auth_validated;
	BOOL faulted_once_before;

	RPC_HDR hdr;
	RPC_HDR_BA hdr_ba;
	RPC_HDR_RB hdr_rb;
	RPC_HDR_REQ hdr_req;

	uint32 key_id; /* security key id */

	uint16 vuid;
	uint16 initial_pipe_vuid;

	int c;			/* socket */

} rpcsrv_struct;

typedef struct ntuser_creds
{
	fstring user_name;
	fstring domain;
	struct pwd_info pwd;

	uint32 ntlmssp_flags;

} CREDS_NT;

struct cli_connection;

typedef void cli_rpc_info;

typedef struct cli_connect_fns
{
	/* create new connection.  strictly speaking, one arg should be
	 * full dce/rpc format: e.g "ncacn_np:\\server\pipe\pipename" */
	cli_rpc_info *(*cli_connect_add)(const char *pipe_name,
			      uint16 vuid,
				  const char *srv_name,
				  const struct ntuser_creds *ntc,
				  BOOL reuse, BOOL *is_new_connection);

	/* terminate client connection */
	void (*cli_connection_free)(cli_rpc_info *con);

	/* get nt creds associated with an msrpc session. */
	struct ntdom_info *(*cli_conn_get_ntinfo)(cli_rpc_info *con);

	/* get a server name associated with a connection */
	const char *(*cli_con_get_srvname)(cli_rpc_info *con);

	/* write a full PDU (in data) to a pipe */
	BOOL (*rpc_api_write)(cli_rpc_info *con, prs_struct *data);

	/* read a full PDU (in rdata) from a pipe */
	BOOL (*rpc_api_rcv_pdu)(struct cli_connection*, cli_rpc_info *con_info,
			prs_struct *rdata);

	/* detect dead servers. The fd is set to -1 when we get an error */
	BOOL (*rpc_con_ok)(cli_rpc_info *con);

	/* write a full PDU (in data) and read a full PDU (in rdata) */
	BOOL (*rpc_api_send_rcv_pdu)(struct cli_connection*,
			cli_rpc_info *con_info,
			prs_struct *data, prs_struct *rdata);

} cli_connect_fns;

typedef struct cli_auth_fns
{
	/* these three will do for now.  they *should* match with server-side */
	BOOL (*create_bind_req) (struct cli_connection *, prs_struct *,
				 uint32, RPC_IFACE *, RPC_IFACE *);
	BOOL (*decode_bind_resp) (struct cli_connection *, prs_struct *);
	BOOL (*create_bind_cont) (struct cli_connection *, prs_struct *,
				  uint32);
	/* creates an authenticated PDU */
	BOOL (*cli_create_pdu) (struct cli_connection *, uint8,
				prs_struct *, int, int *,
				prs_struct *, uint8 *);
	/* decodes an authenticated PDU */
	BOOL (*cli_decode_pdu) (struct cli_connection *, prs_struct *,
				int, int);

} cli_auth_fns;

typedef struct srv_auth_fns
{
	BOOL (*api_is_auth) (RPC_HDR_AUTH *, void **auth_info);

	/* state-based authentication: one to decode, one to generate */
	BOOL (*api_auth_chk) (rpcsrv_struct *, enum RPC_PKT_TYPE);
	BOOL (*api_auth_gen) (rpcsrv_struct *, prs_struct *,
			      enum RPC_PKT_TYPE);

	/* decodes an authenticated PDU */
	BOOL (*api_decode_pdu) (rpcsrv_struct *);
	/* creates an authenticated PDU */
	BOOL (*api_create_pdu) (rpcsrv_struct *, uint32, prs_struct *);

} srv_auth_fns;

struct msrpc_local
{
	fstring srv_name;
	fstring pipe_name;
	struct ntdom_info nt;
#if 0
	cli_auth_fns *auth;
	void *auth_info;
#endif

	int fd;
	BOOL initialised;
	char *inbuf;
	char *outbuf;
};

typedef struct pipes_struct
{
	struct pipes_struct *next, *prev;
	int pnum;
	uint16 vuid;
	uint16 device_state;
	uint16 priority;
	fstring name;
	fstring pipe_srv_name;

	/* remote, server-side rpc redirection */
	struct msrpc_local *m;

	/* local, server-side rpc state processing */
	rpcsrv_struct *l;

}
pipes_struct;

typedef struct msrpc_service_fns
{
	void (*auth_init) (rpcsrv_struct *);
	void (*service_init) (char *);
	BOOL (*reload_services) (BOOL);
	int (*main_init) (int, char *[]);
	void (*idle) (void);

}
msrpc_service_fns;

struct api_struct
{
	const char *name;
	uint16 opnum;
	BOOL (*fn) (PRS_HANDLE *, prs_struct *, prs_struct *);
};

struct acct_info
{
	fstring acct_name;	/* account name */
	fstring acct_desc;	/* account description */
	uint32 rid;		/* domain-relative RID */
};

/*
 * higher order functions for use with msrpc client code
 */

#define ALIAS_FN(fn)\
	void (*fn)(const char*, const DOM_SID*, uint32, const char*)
#define ALIAS_INFO_FN(fn)\
	void (*fn)(const char*, const DOM_SID*, uint32, ALIAS_INFO_CTR *const)
#define ALIAS_MEM_FN(fn)\
	void(*fn)(const char*, const DOM_SID*, uint32, const char*,\
	          uint32, DOM_SID *const *const, char *const *const,\
	          uint32*const)

#define GROUP_FN(fn)\
	void (*fn)(const char*, const DOM_SID*, uint32, const char*)
#define GROUP_INFO_FN(fn)\
	void (*fn)(const char*, const DOM_SID*, uint32, GROUP_INFO_CTR *const)
#define GROUP_MEM_FN(fn)\
	void(*fn)(const char*, const DOM_SID*, uint32, const char*,\
	          uint32, const uint32*, char *const *const,\
	          uint32*const)

#define DOMAIN_FN(fn)\
	void (*fn)(const char*)
#define DOMAIN_INFO_FN(fn)\
	void (*fn)(const char*, const DOM_SID *, uint32, SAM_UNK_CTR *)

#define USER_FN(fn)\
	void (*fn)(const char*, const DOM_SID*, uint32, const char*)
#define USER_INFO_FN(fn)\
	void (*fn)(const char*, const DOM_SID*, uint32, \
	           SAM_USERINFO_CTR *const)
#define USER_MEM_FN(fn)\
	void (*fn)(const char*, const DOM_SID*, uint32, const char*,\
	           uint32, const uint32*, char *const *const, uint32* const)

#define DISP_FN(fn)\
	void (*fn)(const char*, const DOM_SID*, uint16, uint32, \
	           SAM_DISPINFO_CTR *)

#define REG_FN(fn)\
	void (*fn)(int, const char *, int)
#define REG_KEY_FN(fn)\
	void (*fn)(const char*, const char*, time_t)
#define REG_VAL_FN(fn)\
	void (*fn)(const char *, const char*, uint32, const BUFFER2 *)

#define SVC_QUERY_FN(fn)\
	void (*fn)(const QUERY_SERVICE_CONFIG *)
#define SVC_INFO_FN(fn)\
	void (*fn)(const ENUM_SRVC_STATUS *)

#define TPRT_INFO_FN(fn)\
	void (*fn)(const SRV_TPRT_INFO_CTR *)

#define PRINT_INFO_FN(fn)\
	void (*fn)(const char*, uint32, uint32, void  *const *const)
#define JOB_INFO_FN(fn)\
	void (*fn)(const char*, const char*, uint32, uint32, void *const *const)

#endif /* _NT_DOMAIN_H */
