/* 
   Unix SMB/Netbios implementation.
   Misc and possibly useless stuff
   Copyright (C) Jeremy Allison 1992-1998
   Copyright (C) Elrond <elrond@samba-tng.org>  2001-2002.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include <errno.h>
#include "libsamba.h"


/****************************************************************************
 protected memcpy that deals with NULL parameters.
 ****************************************************************************/
BOOL memcpy_zero(void *to, const void *from, size_t size)
{
	if (to == NULL)
	{
		return False;
	}

	if (size == 0)
	{
		DEBUG(5, ("memcpy_zero: Ignoring request to copy 0 bytes\n"));
		return True;
	}

	if (size > 16000000)
	{
		DEBUG(2, ("memcpy_zero: %u bytes, wee...\n", size));
	}

	if (from == NULL)
	{
		memset(to, 0, size);
		return False;
	}

	memcpy(to, from, size);
	return True;
}

/*********************************************************
 For rename across filesystems Patch from Warren Birnbaum 
 <warrenb@hpcvscdp.cv.hp.com>
**********************************************************/

static int copy_reg(char *source, const char *dest)
{
  SMB_STRUCT_STAT source_stats;
  int ifd;
  int ofd;
  char *buf;
  int len;                      /* Number of bytes read into `buf'. */

  sys_lstat (source, &source_stats);
  if (!S_ISREG (source_stats.st_mode))
    return 1;

  if (unlink (dest) && errno != ENOENT)
    return 1;

  if((ifd = sys_open (source, O_RDONLY, 0)) < 0)
    return 1;

  if((ofd = sys_open (dest, O_WRONLY | O_CREAT | O_TRUNC, 0600)) < 0 )
  {
    close (ifd);
    return 1;
  }

  if((buf = malloc( COPYBUF_SIZE )) == NULL)
  {
    close (ifd);  
    close (ofd);  
    unlink (dest);
    return 1;
  }

  while ((len = read(ifd, buf, COPYBUF_SIZE)) > 0)
  {
    if (write_data(ofd, buf, len) < 0)
    {
      close (ifd);
      close (ofd);
      unlink (dest);
      free(buf);
      return 1;
    }
  }
  free(buf);
  if (len < 0)
  {
    close (ifd);
    close (ofd);
    unlink (dest);
    return 1;
  }

  if (close (ifd) < 0)
  {
    close (ofd);
    return 1;
  }
  if (close (ofd) < 0)
    return 1;

  /* chown turns off set[ug]id bits for non-root,
     so do the chmod last.  */

  /* Try to copy the old file's modtime and access time.  */
  {
    struct utimbuf tv;

    tv.actime = source_stats.st_atime;
    tv.modtime = source_stats.st_mtime;
    if (utime (dest, &tv))
      return 1;
  }

  /* Try to preserve ownership.  For non-root it might fail, but that's ok.
     But root probably wants to know, e.g. if NFS disallows it.  */
  if (chown (dest, source_stats.st_uid, source_stats.st_gid)
      && (errno != EPERM))
    return 1;

  if (chmod (dest, source_stats.st_mode & 07777))
    return 1;

  unlink (source);
  return 0;
}

/*******************************************************************
  rename a unix file
********************************************************************/
int file_rename(char *from, char *to)
{
	int rcode = rename(from, to);

	if (errno == EXDEV)
	{
		/* Rename across filesystems needed. */
		rcode = copy_reg(from, to);
	}
	return rcode;
}

/*******************************************************************
prepare to dump a core file - carefully!
********************************************************************/
BOOL dump_core (void)
{
#if DUMP_CORE
	char *p;
	pstring dname;
	pstrcpy (dname, debugf);
	if ((p=strrchr (dname, '/'))) *p=0;
	pstrcat (dname, "/corefiles");
	mkdir (dname, 0700);
	sys_chown (dname, getuid(), getgid());
	chmod (dname, 0700);
	if (chdir (dname))
		return False;
	umask (~(0700));

#ifdef HAVE_GETRLIMIT
#ifdef RLIMIT_CORE
	{
		struct rlimit rlp;
		getrlimit (RLIMIT_CORE, &rlp);
		rlp.rlim_cur = MAX(4*1024*1024,rlp.rlim_cur);
		setrlimit (RLIMIT_CORE, &rlp);
		getrlimit (RLIMIT_CORE, &rlp);
		DEBUG(3,("Core limits now %d %d\n",
			 (int)rlp.rlim_cur, (int)rlp.rlim_max));
	}
#endif
#endif


	DEBUG(0,("Dumping core in %s\n",dname));
	abort ();
	return True;
#else
	return False;
#endif
}


/****************************************************************************
a wrapper for gethostbyname() that tries with all lower and all upper case 
if the initial name fails
****************************************************************************/
struct hostent *Get_Hostbyname(const char *name)
{
	char *name2 = strdup(name);
	struct hostent *ret;

	if (!name2)
	{
		DEBUG(0,
		      ("Memory allocation error in Get_Hostbyname! panic\n"));
		exit(0);
	}


	/* 
	 * This next test is redundent and causes some systems (with
	 * broken isalnum() calls) problems.
	 * JRA.
	 */

#if 0
	if (!isalnum(*name2))
	{
		free(name2);
		return (NULL);
	}
#endif /* 0 */

	ret = sys_gethostbyname(name2);
	if (ret != NULL)
	{
		free(name2);
		return (ret);
	}

	/* try with all lowercase */
	strlower(name2);
	ret = sys_gethostbyname(name2);
	if (ret != NULL)
	{
		free(name2);
		return (ret);
	}

	/* try with all uppercase */
	strupper(name2);
	ret = sys_gethostbyname(name2);
	if (ret != NULL)
	{
		free(name2);
		return (ret);
	}

	/* nothing works :-( */
	free(name2);
	return (NULL);
}


/*******************************************************************
 Reset the 'done' variables so after a client process is created
 from a fork call these calls will be re-done. This should be
 expanded if more variables need reseting.
 ******************************************************************/

void reset_globals_after_fork(void)
{
	/*
	 * Re-seed the random crypto generator, so all smbd's
	 * started from the same parent won't generate the same
	 * sequence.
	 */
	{
		unsigned char dummy;
		generate_random_buffer(&dummy, 1, True);
	}
}


/* needed by the next 2 functions, and by rpcclient/cmd_samr.c */
struct field_info acb_bitfield[] = {
	/* descriptions lifted from <smb.h> */
	{ ACB_PWNOTREQ,  "N", "User password not required" },
	{ ACB_DISABLED,  "D", "User account disabled" },
	{ ACB_HOMDIRREQ, "H", "Home directory required" },
	{ ACB_TEMPDUP,   "T", "Temporary duplicate account" },
	{ ACB_NORMAL,    "U", "Normal user account" },
	{ ACB_MNS,       "M", "MNS logon user account" },
	{ ACB_WSTRUST,   "W", "Workstation trust account" },
	{ ACB_SVRTRUST,  "S", "Server trust account" },
	{ ACB_AUTOLOCK,  "L", "Account locked" },
	{ ACB_PWNOEXP,   "X", "User password does not expire" },
	{ ACB_DOMTRUST,  "I", "Interdomain trust account" },
	{ ACB_PWLOCK,    "P", "Password is locked and cannot be changed remotely" },
	{ 0,             " ", NULL },
	{ 0, NULL, NULL }
};

/**********************************************************
 Encode the account control bits into a string.
 length = length of string to encode into (including terminating
 null). length *MUST BE MORE THAN 2* !
 **********************************************************/

char *pwdb_encode_acct_ctrl(uint16 acct_ctrl, size_t length)
{
	static fstring acct_str;
	size_t i;

	if (length > sizeof(acct_str))
	{
		DEBUG(0, ("BUG: pwdb_encode_acct_ctrl called with length=%d\n",
			  length));
		length = sizeof(acct_str);
	}

	acct_str[0] = '[';
	acct_str[1] = '\0';
	fstrcat (acct_str, bit_field_to_str (acct_ctrl, acb_bitfield, NULL));

	length -= 2;

	i = strlen(acct_str);
	if (i < length)
	{
		memset (acct_str+i, ' ', length-i);
	}

	i = length;

	acct_str[i++] = ']';
	acct_str[i++] = '\0';

	return acct_str;
}

/**********************************************************
 Decode the account control bits from a string.
 **********************************************************/

uint16 pwdb_decode_acct_ctrl(const char *p)
{
	/*
	 * Check if the account type bits have been encoded after the
	 * NT password (in the form [NDHTUWSLXI]).
	 */

	if (*p != '[')
		return 0;
	return str_to_bit_field (p+1, acb_bitfield, NULL);
}

/*************************************************************
 Routine to set 32 hex password characters from a 16 byte array.
**************************************************************/
void pwdb_sethexpwd(char *p, const uchar * pwd, uint16 acct_ctrl)
{
	if (pwd != NULL)
	{
		int i;
		for (i = 0; i < 16; i++)
		{
			slprintf(&p[i * 2], 33, "%02X", pwd[i]);
		}
	}
	else
	{
		if (IS_BITS_SET_ALL(acct_ctrl, ACB_PWNOTREQ))
		{
			safe_strcpy(p, "NO PASSWORDXXXXXXXXXXXXXXXXXXXXX",
				    33);
		}
		else
		{
			safe_strcpy(p, "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
				    33);
		}
	}
}

/*************************************************************
 Routine to get the 32 hex characters and turn them
 into a 16 byte array.
**************************************************************/
BOOL pwdb_gethexpwd(const char *p, char *pwd, uint32 *acct_ctrl)
{
	if (strnequal(p, "NO PASSWORDXXXXXXXXXXXXXXXXXXXXX", 32))
	{
		if (acct_ctrl != NULL)
		{
			*acct_ctrl |= ACB_PWNOTREQ;
		}
		pwd[0] = 0;
		return True;
	}
	else if (strnequal(p, "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX", 32))
	{
		pwd[0] = 0;
		return True;
	}
	else
	{
		return strhex_to_str(pwd, 32, p) == 16;
	}
}


static struct field_info sid_name_info[] = {
	{SID_NAME_UNKNOWN, "UNKNOWN", NULL},	/* default */
	{SID_NAME_USER, "User", NULL},
	{SID_NAME_DOM_GRP, "Domain Group", NULL},
	{SID_NAME_DOMAIN, "Domain", NULL},
	{SID_NAME_ALIAS, "Local Group", NULL},
	{SID_NAME_WKN_GRP, "Well-known Group", NULL},
	{SID_NAME_DELETED, "Deleted", NULL},
	{SID_NAME_INVALID, "Invalid", NULL},
	{0, NULL, NULL}
};

/****************************************************************************
convert a SID_NAME_USE to a string 
****************************************************************************/
const char *get_sid_name_use_str(uint32 sid_name_use)
{
	return enum_field_to_str(sid_name_use, sid_name_info, True);
}
