/* 
   Unix SMB/Netbios implementation.
   Version 2.0.
   LDAP protocol helper functions for SAMBA
   Copyright (C) Jean Franois Micouleau 1998
   Copyright (C) Matthew Chapman 1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
   
*/

#include "includes.h"
#include <errno.h>
#include "libsamba.h"
#include "samdb.h"

#ifdef USE_PASSDB_LDAP

#include <lber.h>
#include <ldap.h>


/* Internal state */
LDAP *ldap_struct = NULL;
LDAPMessage *ldap_results;
LDAPMessage *ldap_entry;

/* LDAP password */
static pstring ldap_secret;


/*******************************************************************
  Open connections to the LDAP server.
 ******************************************************************/	

BOOL ldap_connect(void)
{
	int err;
	char *uri;

	/* XXX Hack so ldap access works only as root - Elrond */
	if (geteuid() != 0) {
		DEBUG(1, ("ldap_connect: Connect denied: euid=%d uid=%d\n",
			  (int) geteuid(), (int) getuid()));
		return False;
	}

	if (ldap_struct) {
		DEBUG(2, ("ldap_connect: Already connected...\n"));
		return True;
	}

	uri = lp_ldap_uri();
	if (uri && *uri) {
#if defined(LDAP_API_FEATURE_X_OPENLDAP) && (LDAP_API_VERSION > 2000)
		DEBUG(0, ("ldap_initialize: %s\n", uri));

		if (ldap_initialize(&ldap_struct, uri) != LDAP_SUCCESS) {
			DEBUG(0, ("ldap_initialize: %s\n", strerror(errno)));
			return (False);
		}
#else
		DEBUG(0, ("ldap_connect: 'ldap uri' requires openldap, v2.0 or higher\n"));
		return False;
#endif
	} else {
		if (!(ldap_struct = ldap_open(lp_ldap_server(), lp_ldap_port()))) {
			DEBUG(0, ("open: %s\n", strerror(errno)));
			return (False);
		}
	}

	err = ldap_simple_bind_s(ldap_struct, lp_ldap_bind_as(), ldap_secret);
	if (err != LDAP_SUCCESS) {
		DEBUG(0, ("bind: %s\n", ldap_err2string(err)));
		return (False);
	}

	DEBUG(2,("Connected to LDAP server\n"));
	return (True);
}

/*******************************************************************
  close connections to the LDAP server.
 ******************************************************************/	

void ldap_disconnect(void)
{
	if(!ldap_struct)
		return;

	if(ldap_results) {
		ldap_msgfree(ldap_results);
		ldap_results = NULL;
	}

	ldap_unbind(ldap_struct);
	ldap_struct = NULL;
	
	DEBUG(2,("Connection closed\n"));
}


/*******************************************************************
  Search the directory using a given filter.
 ******************************************************************/	

BOOL ldap_search_for(char *filter)
{
	int err;

	DEBUG(2,("Searching in [%s] for [%s] with scope [%d]\n", lp_ldap_suffix(), filter, lp_ldap_scope()));

	if(ldap_results) {
		ldap_msgfree(ldap_results);
		ldap_results = NULL;
		ldap_entry = NULL;
	}

	err = ldap_search_s(ldap_struct, lp_ldap_suffix(), lp_ldap_scope(),
			  filter, NULL, 0, &ldap_results);

	if(err != LDAP_SUCCESS) {
		DEBUG(0, ("ERROR: ldap_search_for [%s]: %s\n",
			  filter, ldap_err2string(err)));
		return False;
	}

	DEBUG(2, ("%d matching entries found\n",
		  ldap_count_entries(ldap_struct, ldap_results)));

	ldap_entry = ldap_first_entry(ldap_struct, ldap_results);
	return (True);
}


/*******************************************************************
  Get the first value of an attribute.
 ******************************************************************/

BOOL ldap_get_attribute(char *attribute, char *value)
{
	char **values;

	value[0] = '\0';

	if (! ldap_entry)
		return False;
	
	if(!(values = ldap_get_values(ldap_struct, ldap_entry, attribute)))
	{
		return (False);
	}

	if (convert_string(CH_UTF8, CH_UNIX,
			   values[0], -1, 
			   value, sizeof(pstring)) <= 0)
	{
		ldap_value_free(values);
		return False;
	}

	ldap_value_free(values);
	DEBUG(3, ("get: [%s] = [%s]\n", attribute, value));
	
	return (True);
}


/************************************************************************
  Adds a modification to a LDAPMod queue.
 ************************************************************************/

void ldap_make_mod(LDAPMod ***modlist,int modop, char *attribute, char *value)
{
	LDAPMod **mods;
	int i;
	int j;

	if (value == NULL
	    || (value[0] == '\0' && modop == LDAP_MOD_ADD)) {
		DEBUG(3, ("skipping empty attribute: [%s]\n", attribute));
		return;
	}
	if (value[0] == '\0') {
		DEBUG(3, ("Removing attribute [%s]\n", attribute));
		/* modop = LDAP_MOD_DELETE; */
		value = NULL;
	}
	else
	{
		DEBUG(3, ("set: [%s] = [%s]\n", attribute, value));
	}
	
	mods = *modlist;
	
	if (mods == NULL) {
		mods = (LDAPMod **)malloc(sizeof(LDAPMod *));
		mods[0] = NULL;
	}
	
	for (i = 0; mods[i] != NULL; ++i) {
		if (mods[i]->mod_op == modop && 
		    !strcasecmp(mods[i]->mod_type, attribute)) {
			break;
		}
	}
	
	if (mods[i] == NULL) {
		mods = (LDAPMod **)realloc(mods, (i+2) * sizeof(LDAPMod *));
		mods[i] = (LDAPMod *)malloc(sizeof(LDAPMod));
		mods[i]->mod_op = modop;
		mods[i]->mod_values = NULL;
		mods[i]->mod_type = strdup(attribute);
		mods[i+1] = NULL;
	}

	j = 0;
	if (mods[i]->mod_values) {
		for (; mods[i]->mod_values[j]; j++);
	}
	if (! value)
		j--;

	mods[i]->mod_values = (char **)realloc(mods[i]->mod_values,
					  (j+2) * sizeof(char *));

	if (value)
	{
		void *utf8_value = NULL;
		if (convert_string_allocate(CH_UNIX, CH_UTF8,
					    value, strlen(value)+1,
					    &utf8_value) == -1)
		{
			utf8_value = strdup(value);
		}
		mods[i]->mod_values[j] = utf8_value;
	}

	mods[i]->mod_values[j+1] = NULL;

	*modlist = mods;
}


/************************************************************************
  Commit changes to a directory entry.
 *************************************************************************/
BOOL ldap_commit_mods(const char *attribute, const char *value,
		      LDAPMod **mods, BOOL add, const char *subcontainer)
{
	pstring filter;
	char *dn;
	int err = 0;
	BOOL rc;

	if (!mods)
	{
		DEBUG(1, ("ldap_commit_mods for object (%s=%s): mods==NULL! "
			  "Doing nothing, returning Success.\n",
			  attribute, value));
		return True;
	}

	slprintf(filter, sizeof(filter)-1, "%s=%s", attribute, value);

	if (!ldap_connect())
		return (False);

	ldap_search_for(filter);

	if (ldap_entry)
	{
		dn = ldap_get_dn(ldap_struct, ldap_entry);
		DEBUG(3, ("ldap_commit_mods: Commiting changes to object "
			  "(%s=%s) dn=%s\n", attribute, value, dn));
		err = ldap_modify_s(ldap_struct, dn, mods);
		free(dn);
	}
	else if (add)
	{
/* Someone please explain - what exactly is 'ldap dn add form', why is it
 * better than 'ldap suffix', and do we need to support it.
 */
#if 0
		char *form = lp_ldap_dn_add_form();
		if (form && *form)
		{
			pstring newDn;
			slprintf(newDn, sizeof(newDn)-1, form, value);
			err = ldap_add_s(ldap_struct, newDn, mods);
		}
		else
#endif
		{
			char *suf = lp_ldap_suffix();

			if (subcontainer && *subcontainer)
			{
				pstring dummy;
				pstrcpy(dummy, subcontainer);
				trim_string(dummy, " ", " ");
				trim_string(dummy, ",", ",");
				pstrcat(filter, ",");
				pstrcat(filter, dummy);
			}

			if (suf && *suf)
			{
				pstrcat(filter, ",");
				pstrcat(filter, suf);
			}
			err = ldap_add_s(ldap_struct, filter, mods);
		}
	}

	if (err == LDAP_SUCCESS)
	{
		DEBUG(2,("Updated entry [%s]\n", value));
		rc = True;
	} else {
		DEBUG(0,("update: %s\n", ldap_err2string(err)));
		rc = False;
	}

	ldap_disconnect();
	ldap_mods_free(mods, 1);
	return rc;
}


/************************************************************************
  Modify user information given an smb_passwd struct.
 *************************************************************************/
static BOOL ldap_addpw(struct smb_passwd *newpwd)
{
	LDAPMod **mods;

	if (!newpwd || !ldap_allocaterid(&newpwd->user_rid))
		return (False);

	ldap_smbpwmods(newpwd, &mods, LDAP_MOD_ADD, False);
	return ldap_commit_mods("uid", newpwd->unix_name, mods, True, NULL)
		&& ldap_incr_seq_num();
}

static BOOL ldap_modpw(struct smb_passwd *pwd, BOOL override)
{
	LDAPMod **mods;

	if (!pwd)
		return (False);

	ldap_smbpwmods(pwd, &mods, LDAP_MOD_REPLACE, False);
	return ldap_commit_mods("uid", pwd->unix_name, mods, False, NULL)
		&& ldap_incr_seq_num();
}

static BOOL ldap_del_user(uint32 user_rid)
{
	char *dn;
	int err;

	if (!lp_ldap_delete_dn())
		return False;

	if (!ldap_connect())
		return (False);

	ldap_search_by_rid(user_rid);

	if (!ldap_entry || !(dn = ldap_get_dn(ldap_struct, ldap_entry)))
	{
		ldap_disconnect();
		return (False);
	}

	err = ldap_delete_s(ldap_struct, dn);
	free(dn);
	ldap_disconnect();

	if (err != LDAP_SUCCESS)
	{
		DEBUG(0, ("delete: %s\n", ldap_err2string(err)));
		return (False);
	}

	return True;
}

static struct smb_passdb_ops ldap_ops =
{
	NULL,

	NULL,
	NULL,
	NULL,
	NULL,

	NULL,
	NULL,
	NULL,
	ldap_addpw,
	ldap_modpw,
	ldap_del_user
};

struct smb_passdb_ops *ldap_initialise_password_db(void)
{
	FILE *pwdfile;
	char *pwdfilename;
	char *p;

	/* 'ldap suffix' is required; if missing, try another db */
	p = lp_ldap_suffix();
	if (!p || !*p)
	{
		return NULL;
	}

	pwdfilename = lp_ldap_passwd_file();

	if(pwdfilename[0]) {
		if((pwdfile = sys_fopen(pwdfilename, "r"))) {
			fgets(ldap_secret, sizeof(ldap_secret), pwdfile);
			if((p = strchr(ldap_secret, '\n')))
				*p = 0;
			fclose(pwdfile);
		} else {
			DEBUG(0,("Failed to open LDAP passwd file\n"));
		}
	}

	return &ldap_ops;
}

#else
struct smb_passdb_ops *ldap_initialise_password_db(void)
{
	return NULL;
}
#endif
