/* 
   Unix SMB/Netbios implementation.
   Version 2.0.
   LDAP protocol helper functions for SAMBA
   Copyright (C) Matthew Chapman 1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
   
*/

#include "includes.h"
#include "libsamba.h"
#include "samdb.h"

#ifdef USE_PASSDB_LDAP

#include <lber.h>
#include <ldap.h>

/* Internal state */
extern LDAP *ldap_struct;
extern LDAPMessage *ldap_results;
extern LDAPMessage *ldap_entry;


/*******************************************************************
  NT name/RID search functions.
 ******************************************************************/

BOOL ldap_search_by_rid(uint32 rid)
{
	fstring filter;

	slprintf(filter, sizeof(filter) - 1,
		 "(&(rid=%x)(objectclass=sambaAccount))", rid);
	return ldap_search_for(filter);
}

static BOOL ldap_search_by_ntname(const char *ntname)
{
        fstring filter;

        slprintf(filter, sizeof(filter)-1,
                 "(&(ntuid=%s)(objectclass=sambaAccount))", ntname);
        return ldap_search_for(filter);
}


/*******************************************************************
  Construct an smb_passwd structure
 ******************************************************************/
static struct smb_passwd *ldap_getpw(void)
{
	static struct smb_passwd smbpw;
	static pstring unix_name;
	static pstring nt_name;
	static unsigned char smblmpwd[16];
	static unsigned char smbntpwd[16];
	pstring temp;

	pwdb_init_smb(&smbpw);

	if(!ldap_entry)
		return NULL;

	if(!ldap_get_attribute("uid", unix_name)) {
		DEBUG(0,("Missing uid\n"));
		return NULL;
	}
	smbpw.unix_name = unix_name;

	DEBUG(2,("Retrieving account [%s]\n",unix_name));

	if(!ldap_get_attribute("ntuid", nt_name)) {
		DEBUG(0,("Missing ntuid\n"));
		return NULL;
	}
	smbpw.nt_name = nt_name;

	if(!ldap_get_attribute("rid", temp)) {
		DEBUG(0,("Missing rid\n"));
		return NULL;
	}
	smbpw.user_rid = strtol(temp, NULL, 16);

	if(ldap_get_attribute("acctFlags", temp))
		smbpw.acct_ctrl = pwdb_decode_acct_ctrl(temp);
	else
		smbpw.acct_ctrl = ACB_NORMAL;

	if(ldap_get_attribute("lmPassword", temp)) {
		pwdb_gethexpwd(temp, smblmpwd, NULL);
		smbpw.smb_passwd = smblmpwd;
	} else {
		smbpw.smb_passwd = NULL;
		smbpw.acct_ctrl |= ACB_DISABLED;
	}

	if(ldap_get_attribute("ntPassword", temp)) {
		pwdb_gethexpwd(temp, smbntpwd, NULL);
		smbpw.smb_nt_passwd = smbntpwd;
	} else {
		smbpw.smb_nt_passwd = NULL;
	}

	if(ldap_get_attribute("pwdLastSet", temp))
		smbpw.pass_last_set_time = (time_t)strtol(temp, NULL, 16);
	else
		smbpw.pass_last_set_time = (time_t)(-1);

	return &smbpw;
}


/************************************************************************
  Queues the necessary modifications to save a smb_passwd structure
 ************************************************************************/

void ldap_smbpwmods(struct smb_passwd *newpwd, LDAPMod ***mods,
		    int operation, BOOL add_base_attr)
{
	fstring temp;
	int i;

	*mods = NULL;

	if(operation == LDAP_MOD_ADD) { /* immutable attributes */
		if (add_base_attr) {
			ldap_make_mod(mods, LDAP_MOD_ADD,
				      "objectclass", "account");
			ldap_make_mod(mods, LDAP_MOD_ADD,
				      "uid", newpwd->unix_name);
		}

		ldap_make_mod(mods, LDAP_MOD_ADD, "objectclass", "sambaAccount");
		ldap_make_mod(mods, LDAP_MOD_ADD, "ntuid", newpwd->nt_name);
		slprintf(temp, sizeof(temp)-1, "%x", newpwd->user_rid);
		ldap_make_mod(mods, LDAP_MOD_ADD, "rid", temp);
	}

	if (newpwd->smb_passwd) {
		for( i = 0; i < 16; i++) {
			slprintf(&temp[2*i], 3, "%02X", newpwd->smb_passwd[i]);
		}
		ldap_make_mod(mods, operation, "lmPassword", temp);
	}

	if (newpwd->smb_nt_passwd) {
		for( i = 0; i < 16; i++) {
			slprintf(&temp[2*i], 3, "%02X", newpwd->smb_nt_passwd[i]);
		}
		ldap_make_mod(mods, operation, "ntPassword", temp);
	}

	slprintf(temp, sizeof(temp)-1, "%08X", newpwd->pass_last_set_time);
	ldap_make_mod(mods, operation, "pwdLastSet", temp);

	ldap_make_mod(mods, operation, "acctFlags",
		      pwdb_encode_acct_ctrl(newpwd->acct_ctrl,
					    NEW_PW_FORMAT_SPACE_PADDED_LEN));
}


/************************************************************************
  Return next available RID, starting from 1000
 ************************************************************************/
BOOL ldap_allocaterid(uint32 *rid)
{
	pstring newdn;
	fstring rid_str;
	LDAPMod **mods;
	char *dn;
	int err;

	DEBUG(2, ("Allocating new RID\n"));

	if (!ldap_connect())
		return (False);

	ldap_search_for("(&(id=root)(objectClass=sambaConfig))");

	if (ldap_entry && ldap_get_attribute("nextrid", rid_str))
		*rid = strtol(rid_str, NULL, 16);
	else
		*rid = 1000;

	mods = NULL;
	if(!ldap_entry)
	{
		ldap_make_mod(&mods, LDAP_MOD_ADD, "objectClass",
			      "sambaConfig");
		ldap_make_mod(&mods, LDAP_MOD_ADD, "id", "root");
	}

	slprintf(rid_str, sizeof(fstring)-1, "%x", (*rid) + 1);
	ldap_make_mod(&mods, LDAP_MOD_REPLACE, "nextrid", rid_str);

	if (ldap_entry)
	{
                dn = ldap_get_dn(ldap_struct, ldap_entry);
                err = ldap_modify_s(ldap_struct, dn, mods);
                free(dn);
	} else {
		pstrcpy(newdn, "id=root, ");
		pstrcat(newdn, lp_ldap_suffix());
		err = ldap_add_s(ldap_struct, newdn, mods);
	}

	ldap_mods_free(mods, 1);

	ldap_disconnect();

	if(err != LDAP_SUCCESS)
	{
		DEBUG(0,("nextrid update: %s\n", ldap_err2string(err)));
		return (False);
	}

	return (True);
}

static uint32 ldap_get_samsequence_number(BOOL close_con)
{
	fstring str;
	uint32 seqnum = 0;

	become_root();

	if (!ldap_connect())
	{
		unbecome_root();
		return seqnum;
	}
	unbecome_root();

	ldap_search_for("(&(id=root)(objectClass=sambaConfig))");

	if (ldap_entry && ldap_get_attribute("samSequenceNumber", str))
		seqnum = strtol(str, NULL, 10);

	if (close_con)
		ldap_disconnect();

	return seqnum;
}

static uint32 ldap_getsamseqnum(void)
{
	return ldap_get_samsequence_number(True);
}

BOOL ldap_incr_seq_num(void)
{
	fstring str;
	LDAPMod **mods;
	char *dn;
	int err;
	uint32 seqnum;

	if (! lp_ldap_sam_seqnum())
		return True;

	seqnum = ldap_get_samsequence_number(False);

	mods = NULL;

	if (seqnum)
	{
		slprintf(str, sizeof(fstring)-1, "%d", seqnum);
		ldap_make_mod(&mods, LDAP_MOD_DELETE, "samSequenceNumber", str);
	}

	seqnum++;

	slprintf(str, sizeof(fstring)-1, "%d", seqnum);
	ldap_make_mod(&mods, LDAP_MOD_ADD, "samSequenceNumber", str);

	dn = ldap_get_dn(ldap_struct, ldap_entry);
	err = ldap_modify_s(ldap_struct, dn, mods);
	free(dn);
	ldap_mods_free(mods, 1);

	ldap_disconnect();

	if(err != LDAP_SUCCESS)
	{
		DEBUG(0,("ldap_incr_seq_num: update: %s\n", ldap_err2string(err)));
		return False;
	}

	return True;
}


/*******************************************************************
  Store NTTIMEs as time_t's.
 ******************************************************************/

static void ldap_save_time(LDAPMod ***modlist, int modop, char *attribute,
                        NTTIME *nttime)
{
        fstring tstr;
        time_t t;

        t = nt_time_to_unix(nttime);

        if(t == -1)
	{
		ldap_make_mod(modlist, modop, attribute, "");
		return;
	}

	slprintf(tstr, sizeof(tstr)-1, "%08lx", t);
        ldap_make_mod(modlist, modop, attribute, tstr);
}

static void ldap_read_time(char *attribute, NTTIME *nttime)
{
        fstring timestr;
        time_t t;

        if(ldap_get_attribute(attribute, timestr))
        {
                t = (time_t)strtol(timestr, NULL, 16);
                unix_to_nt_time(nttime, t);
        }
}


/*******************************************************************
  Contruct a sam_passwd structure.
 ******************************************************************/

static struct sam_passwd *ldapsam_getsam(void)
{
	static pstring full_name;
	static pstring acct_desc;
	static pstring home_dir;
	static pstring home_drive;
	static pstring logon_script;
	static pstring profile_path;
	static pstring workstations;
	pstring temp;
	struct sam_passwd *sam21;
	struct smb_passwd *smbpw;

	/* to obtain default values from smb.conf [mami@mami.net, 2001-03-22] */
	user_struct bogus_user_struct;

	if(!ldap_entry)
		return NULL;

	ZERO_STRUCT(bogus_user_struct);

	smbpw = ldap_getpw();
	sam21 = pwdb_smb_to_sam(smbpw);	

	/* [gavin.bravery@lineone.net 2001-04-19]
	   HACK to make %U work in substitutions below
	*/
	fstrcpy(bogus_user_struct.requested_name , smbpw->nt_name);
	fstrcpy(bogus_user_struct.unix_name      , smbpw->unix_name);

	if(ldap_get_attribute("grouprid", temp))
		sam21->group_rid = strtol(temp, NULL, 16);
	
	if(ldap_get_attribute("cn", full_name))
		sam21->full_name = full_name;

	if(ldap_get_attribute("description", acct_desc))
		sam21->acct_desc = acct_desc;

	if(ldap_get_attribute("smbHome", home_dir))
		sam21->home_dir = home_dir;
	else
		sam21->home_dir = lp_logon_home(&bogus_user_struct);

	if(ldap_get_attribute("homeDrive", home_drive))
		sam21->dir_drive = home_drive;
	else
		sam21->dir_drive = lp_logon_drive(&bogus_user_struct);

	if(ldap_get_attribute("script", logon_script))
		sam21->logon_script = logon_script;
	else
		sam21->logon_script = lp_logon_script(&bogus_user_struct);

	if(ldap_get_attribute("profile", profile_path))
		sam21->profile_path = profile_path;
	else
		sam21->profile_path = lp_logon_path(&bogus_user_struct);

	if(ldap_get_attribute("workstations", workstations))
		sam21->workstations = workstations;

	ldap_read_time("pwdCanChange", &sam21->pass_can_change_time);
	ldap_read_time("pwdMustChange", &sam21->pass_must_change_time);
	ldap_read_time("logonTime", &sam21->logon_time);
	ldap_read_time("logoffTime", &sam21->logoff_time);
	ldap_read_time("kickoffTime", &sam21->kickoff_time);

        sam21->logon_divs = 168; /* hours per week */
        sam21->hours_len = 21; /* 21 times 8 bits = 168 */
        memset(sam21->hours, 0xff, sam21->hours_len); /* all hours */
        sam21->unknown_str = NULL;
        sam21->munged_dial = NULL;

	standard_sub_vuser(logon_script, NULL);
	standard_sub_vuser(profile_path, NULL);
	standard_sub_vuser(home_drive, NULL);
	standard_sub_vuser(home_dir, NULL);
	standard_sub_vuser(workstations, NULL);

        ldap_entry = ldap_next_entry(ldap_struct, ldap_entry);
	return sam21;
}


/************************************************************************
  Queues the necessary modifications to save a sam_passwd structure
 ************************************************************************/

static void ldapsam_sammods(struct sam_passwd *newpwd, LDAPMod ***mods,
			   int operation, BOOL add_base_attr)
{
	struct smb_passwd *smbpw;
	pstring temp;

	smbpw = pwdb_sam_to_smb(newpwd);
	ldap_smbpwmods(smbpw, mods, operation, add_base_attr);

	slprintf(temp, sizeof(temp)-1, "%x", newpwd->group_rid);
	ldap_make_mod(mods, operation, "grouprid", temp);

	ldap_make_mod(mods, operation, "cn", newpwd->full_name);
	ldap_make_mod(mods, operation, "description", newpwd->acct_desc);
	ldap_make_mod(mods, operation, "smbHome", newpwd->home_dir);
	ldap_make_mod(mods, operation, "homeDrive", newpwd->dir_drive);
	ldap_make_mod(mods, operation, "script", newpwd->logon_script);
	ldap_make_mod(mods, operation, "profile", newpwd->profile_path);
	ldap_make_mod(mods, operation, "workstations", newpwd->workstations);

	ldap_save_time(mods, operation, "pwdCanChange",
			&newpwd->pass_can_change_time);
	ldap_save_time(mods, operation, "pwdMustChange",
			&newpwd->pass_must_change_time);
	ldap_save_time(mods, operation, "logonTime",
			&newpwd->logon_time);
	ldap_save_time(mods, operation, "logoffTime",
			&newpwd->logoff_time);
	ldap_save_time(mods, operation, "kickoffTime",
			&newpwd->kickoff_time);
}


/***************************************************************
  Begin/end account enumeration.
 ****************************************************************/

static void *ldapsam_enumfirst(BOOL update)
{
	if (!ldap_connect())
		return NULL;

	ldap_search_for("objectclass=sambaAccount");

	return ldap_struct;
}

static void ldapsam_enumclose(void *vp)
{
	ldap_disconnect();
}


/*************************************************************************
  Return sam_passwd information.
 *************************************************************************/

static struct sam_passwd *ldapsam_getsambynam(const char *name)
{
	struct sam_passwd *ret;

	if(!ldap_connect())
		return NULL;

	ldap_search_by_ntname(name);
	ret = ldapsam_getsam();

	ldap_disconnect();
	return ret;
}

static struct sam_passwd *ldapsam_getsambyrid(uint32 user_rid)
{
	struct sam_passwd *ret;

	if(!ldap_connect())
	   return NULL;

	ldap_search_by_rid(user_rid);
	ret = ldapsam_getsam();

	ldap_disconnect();
	return ret;
}

static struct sam_passwd *ldapsam_getcurrentsam(void *vp)
{
        return ldapsam_getsam();
}


/************************************************************************
  Modify user information given a sam_passwd struct.
 *************************************************************************/

static BOOL ldapsam_addsam(struct sam_passwd *newpwd)
{	
	LDAPMod **mods;
	const char *subcontext;
	pstring filter;
	BOOL add_base;

	if (!newpwd || !ldap_allocaterid(&newpwd->user_rid))
		return (False);

	if (! ldap_connect())
		return False;

	slprintf(filter, sizeof(filter)-1, "uid=%s", newpwd->unix_name);

	ldap_search_for(filter);

	if (ldap_entry)
		add_base = False;
	else
		add_base = True;

	ldapsam_sammods(newpwd, &mods, LDAP_MOD_ADD, add_base);
	if (newpwd->unix_name[strlen(newpwd->unix_name) - 1] == '$')
		subcontext = lp_ldap_computers_subcontext();
	else
		subcontext = lp_ldap_users_subcontext();

	return ldap_commit_mods("uid", newpwd->unix_name, mods, True,
				subcontext)
		&& ldap_incr_seq_num();
}

static BOOL ldapsam_modsam(struct sam_passwd *pwd, BOOL override)
{
	LDAPMod **mods;

	if (!pwd)
		return (False);

	ldapsam_sammods(pwd, &mods, LDAP_MOD_REPLACE, False);
	return ldap_commit_mods("uid", pwd->unix_name, mods, False, NULL)
		&& ldap_incr_seq_num();
}



static struct sam_passdb_ops ldapsam_ops =
{
	ldapsam_enumfirst,
	ldapsam_enumclose,
	NULL,
	NULL,

	ldapsam_getsambynam,
	NULL,
	ldapsam_getsambyrid,
	ldapsam_getcurrentsam,
	ldapsam_addsam,
	ldapsam_modsam,
	NULL,
	NULL
};

struct sam_passdb_ops *ldap_initialise_sam_password_db(void)
{
	/* 'ldap suffix' is required; if missing, try another db */
	char *p = lp_ldap_suffix();
	if (!p || !*p)
	{
		return NULL;
	}

	if (lp_ldap_sam_seqnum())
		ldapsam_ops.getsamseqnum = ldap_getsamseqnum;

	return &ldapsam_ops;
}

#else
struct sam_passdb_ops *ldap_initialise_sam_password_db(void)
{
	return NULL;
}
#endif
