/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   DCE/RPC client ncacn_np functions
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "libsamba.h"
#include "byteorder.h"
#include "libsmb.h"
#include "doserr.h"
#include "rpc_parse.h"
#include "rpc_client.h"


/*******************************************************************
 uses SMBreadX to get rest of rpc data
 ********************************************************************/

static BOOL rpc_read(struct cli_state *cli, uint16 fnum,
		     prs_struct *rdata, uint32 data_to_read,
		     uint32 rdata_offset, BOOL one_only)
{
	size_t size = cli->nt.max_recv_frag;
	int file_offset = 0;
	int num_read;
	char *data;
	uint32 new_data_size = rdata_offset + data_to_read;
	uint8 cls;
	uint32 type;

	DEBUG(5,
	      ("rpc_read: data_to_read: %d data offset: %d file offset: %d\n",
	       data_to_read, rdata_offset, file_offset));

	if (new_data_size > prs_data_size(rdata))
	{
		prs_grow_data(rdata, True, new_data_size, True);
		DEBUG(5, ("rpc_read: grow buffer to %d\n", prs_data_size(rdata)));
	}

	data = rdata->data + rdata_offset;

	do			/* read data using SMBreadX */
	{
		if (size > data_to_read)
		{
			size = data_to_read;
		}

		num_read = cli_read_one(cli, fnum, data, file_offset, size);

		DEBUG(5, ("rpc_read: read offset: %d read: %d to read: %d\n",
			  file_offset, num_read, data_to_read));

		data_to_read -= num_read;
		file_offset += num_read;
		data += num_read;

		if (cli_dos_error(cli, &cls, &type))
		{
			if (cls != ERRDOS || type != ERRmoredata)
			{
				return False;
			}
		}

	}
	while (!one_only && num_read > 0 && data_to_read > 0);

	rdata->end = new_data_size;

	DEBUG(5, ("rpc_read: offset end: 0x%x.  data left to read:0x%x\n",
		  rdata->end, data_to_read));

	return True;
}

/****************************************************************************
 send data on an rpc pipe, which *must* be in one fragment.
 receive response data from an rpc pipe, which may be large...

 read the first fragment: unfortunately have to use SMBtrans for the first
 bit, then SMBreadX for subsequent bits.

 if first fragment received also wasn't the last fragment, continue
 getting fragments until we _do_ receive the last fragment.

 [note: from a data abstraction viewpoint, this function is marginally
        complicated by the return side of cli_api_pipe getting in the way
        (i.e, the SMB header stuff).  the proper way to do this is to split
        cli_api_pipe down into receive / transmit.  oh, and split cli_readx
        down.  in other words, state-based (kernel) techniques...]

 ****************************************************************************/

static BOOL cli_send_trans_data(struct cli_state *cli,
				uint16 fnum, prs_struct *data,
				int max_data_len, prs_struct *rdata)
{
	uint16 cmd = 0x0026;
	uint16 setup[2];	/* only need 2 uint16 setup parameters */
	char *rparam = NULL;
	uint32 rparam_len = 0;
	/*
	 * Setup the pointers to the outgoing.
	 */
	char *rdata_t = NULL;
	uint32 rdata_len = 0;
	char *pipe_name = "\\PIPE\\\0\0\0";
	int pipe_len = 8;
	int setup_len = 2;
	/*
	 * Setup the pointers from the incoming.
	 */
	char *pdata = prs_data(data, 0);
	int data_len = data ? prs_data_size(data) : 0;
	data_len = MIN(max_data_len, data_len);
	/* create setup parameters. */
	setup[0] = cmd;
	setup[1] = fnum;	/* pipe file handle.  got this from an SMBOpenX. */
	DEBUG(5,
	      ("cli_send_trans_data: data_len: %d cmd:%x fnum:%x\n",
	       data_len, cmd, fnum));
	/* send the data: receive a response. */
	if (!cli_api_pipe(cli, pipe_name, pipe_len, setup, setup_len, 0,	/* Setup, length, max */
			  NULL, 0, 0,	/* Params, length, max */
			  pdata, data_len, max_data_len,	/* data, length, max */
			  &rparam, &rparam_len,	/* return param, length */
			  &rdata_t, &rdata_len))	/* return data, len */
	{
		fstring errstr;
		cli_safe_errstr(cli, errstr, sizeof(errstr) - 1);
		DEBUG(0,
		      ("cli_pipe: return critical error. Error was %s\n",
		       errstr));
		return False;
	}

	safe_free(rparam);
	if (rdata_len != 0)
	{
		BOOL ret = prs_append_data(rdata, rdata_t, rdata_len);
		safe_free(rdata_t);
		return ret;
	}

	return True;
}

/****************************************************************************
 send data on an rpc pipe, which *must* be in one fragment.
 receive response data from an rpc pipe, which may be large...
 ****************************************************************************/
static BOOL cli_send_and_rcv_pdu_trans(struct cli_connection *con,
				       struct cli_state *cli,
				       uint16 fnum, prs_struct *data,
				       prs_struct *rdata, int max_send_pdu)
{
	int len;
	cli_auth_fns *auth = cli_conn_get_authfns(con);
	BOOL first = True;
	BOOL last = True;
	RPC_HDR rhdr;
	size_t data_len = prs_data_size(data);
	int max_data_len = MAX(data_len, 2048);
	DEBUG(5, ("cli_send_and_rcv_pdu_trans: fnum:%x\n", fnum));
	DEBUG(10, ("cli_send_and_rcv_pdu_trans: len: %d\n", data_len));
	if (!cli_send_trans_data(cli, fnum, data, max_data_len, rdata))
	{
		return False;
	}

	if (rdata->data == NULL)
		return False;
	/**** parse the header: check it's a response record */
	rdata->start = 0;
	rdata->end = prs_data_size(rdata);
	rdata->offset = 0;
	if (!rpc_check_hdr(rdata, &rhdr, &first, &last, &len))
	{
		return False;
	}

	prs_set_packtype(rdata, rhdr.pack_type);
	if (rhdr.pkt_type == RPC_BINDACK)
	{
		if (!last && !first)
		{
			DEBUG(5,
			      ("cli_pipe: bug in AS/U, setting fragment first/last ON\n"));
			first = True;
			last = True;
		}
	}


	DEBUG(5, ("cli_pipe: len left: %d smbtrans read: %d\n",
		  len, prs_data_size(rdata)));
	/* check if data to be sent back was too large for one SMB. */
	/* err status is only informational: the _real_ check is on the length */
	if (len > 0)		/* || err == (0x80000000 | STATUS_BUFFER_OVERFLOW)) */
	{
		if (!rpc_read(cli, fnum, rdata, len, prs_data_size(rdata), False))
		{
			return False;
		}

		if (rhdr.auth_len != 0 &&
		    (auth->cli_decode_pdu == NULL ||
		     !auth->cli_decode_pdu(con, rdata,
					   rhdr.frag_len, rhdr.auth_len)))
		{
			return False;
		}

	}

	return True;
}

/****************************************************************************
 send data on an rpc pipe, which *must* be in one fragment.
 receive response data from an rpc pipe, which may be large...
 ****************************************************************************/

static BOOL cli_send_and_rcv_pdu_rw(struct cli_connection *con,
			     struct cli_state *cli,
			     uint16 fnum, prs_struct *data,
			     prs_struct *rdata, int max_send_pdu)
{
	int len;
	int data_offset = 0;
	cli_auth_fns *auth = cli_conn_get_authfns(con);
	BOOL first = True;
	BOOL last = True;
	RPC_HDR rhdr;
	int max_data_len = 2048;
	int write_mode = 0x000c;
	char *d = NULL;
	size_t data_left = prs_data_size(data);
	size_t data_len = prs_data_size(data);
	DEBUG(5, ("cli_send_and_rcv_pdu_rw: fnum:%x\n", fnum));
	while (data_offset < data_len)
	{
		DEBUG(10,
		      ("cli_send_and_rcv_pdu_rw: off: %d len: %d left: %d\n",
		       data_offset, data_len, data_left));
		if (d == NULL)
		{
			d = (char *)malloc(data_left + 2);
			if (d == NULL)
			{
				return False;
			}
			SSVAL(d, 0, data_len);
			memcpy(d + 2, data->data, data_len);
			data_len += 2;
		}
		max_data_len = MIN(max_data_len, data_len - data_offset);
		if (cli_write
		    (cli, fnum, write_mode, d, data_offset,
		     max_data_len, data_left) != max_data_len)
		{
			return False;
		}
		write_mode = 0x0004;
		d += max_data_len;
		data_offset += max_data_len;
		data_left -= max_data_len;
	}
	if (!rpc_read(cli, fnum, rdata, max_send_pdu, 0, True))
	{
		return False;
	}

	if (rdata->data == NULL)
		return False;
	/**** parse the header: check it's a response record */
	rdata->start = 0;
	rdata->end = prs_data_size(rdata);
	rdata->offset = 0;
	if (!rpc_check_hdr(rdata, &rhdr, &first, &last, &len))
	{
		return False;
	}

	prs_set_packtype(rdata, rhdr.pack_type);
	if (rhdr.pkt_type == RPC_BINDACK)
	{
		if (!last && !first)
		{
			DEBUG(5,
			      ("cli_pipe: bug in AS/U, setting fragment first/last ON\n"));
			first = True;
			last = True;
		}
	}

	if (rhdr.pkt_type == RPC_RESPONSE)
	{
		RPC_HDR_RESP rhdr_resp;
		smb_io_rpc_hdr_resp("rpc_hdr_resp", &rhdr_resp, rdata, 0);
	}

	DEBUG(5, ("cli_pipe: len left: %d smbtrans read: %d\n",
		  len, prs_data_size(rdata)));
	/* check if data to be sent back was too large for one SMB. */
	/* err status is only informational: the _real_ check is on the length */
	if (len > 0)
	{
		if (!rpc_read(cli, fnum, rdata, len, prs_data_size(rdata), False))
		{
			return False;
		}
	}

	if (rhdr.auth_len != 0 &&
	    (auth->cli_decode_pdu == NULL ||
	     !auth->cli_decode_pdu(con, rdata, rhdr.frag_len, rhdr.auth_len)))
	{
		return False;
	}

	return True;
}

/****************************************************************************
 send data on an rpc pipe, which *must* be in one fragment.
 receive response data from an rpc pipe, which may be large...
 ****************************************************************************/
static BOOL cli_send_and_rcv_pdu(struct cli_connection *con,
			  struct cli_state *cli, uint16 fnum,
			  prs_struct *data, prs_struct *rdata,
			  int max_send_pdu)
{
	if (True)
	{
		return cli_send_and_rcv_pdu_trans(con, cli, fnum,
						  data, rdata, max_send_pdu);
	}
	else
	{
		return cli_send_and_rcv_pdu_rw(con, cli, fnum, data,
					       rdata, max_send_pdu);
	}
}


static BOOL cli_rcv_pdu(struct cli_connection *con,
			struct cli_state *cli, uint16 fnum, prs_struct *rdata)
{
	RPC_HDR_RESP rhdr_resp;
	RPC_HDR rhdr;
	char readbuf[0x19];
	int num_read;
	BOOL first = True;
	BOOL last = True;
	int len;
	cli_auth_fns *auth = cli_conn_get_authfns(con);

	/* with a little help by Scummer */
	num_read = cli_read_one(cli, fnum, readbuf, 0, 0x18);
	DEBUG(5, ("cli_pipe: read header (size:%d)\n", num_read));
	if (num_read != 0x18)
		return False;
	prs_append_data(rdata, readbuf, num_read);
	if (!rpc_check_hdr(rdata, &rhdr, &first, &last, &len))
	{
		return False;
	}

	prs_set_packtype(rdata, rhdr.pack_type);
	smb_io_rpc_hdr_resp("rpc_hdr_resp", &rhdr_resp, rdata, 0);
	if (!rpc_read(cli, fnum, rdata, len, prs_data_size(rdata), False))
	{
		return False;
	}

	if (rhdr.auth_len != 0 &&
	    (auth->cli_decode_pdu == NULL ||
	     !auth->cli_decode_pdu(con, rdata, rhdr.frag_len, rhdr.auth_len)))
	{
		return False;
	}

	return True;
}

/****************************************************************************
 set ntlmssp negotiation flags
 ****************************************************************************/

static void cli_nt_set_ntlmssp_flgs(struct cli_state *cli, uint32 ntlmssp_flgs)
{
	cli->nt.ntlmssp_cli_flgs = ntlmssp_flgs;
}


/****************************************************************************
 open a session
 ****************************************************************************/

BOOL cli_nt_session_open(struct cli_state *cli,
			 const char *pipe_name, uint16 *fnum)
{
	/******************* open the pipe *****************/
	if (IS_BITS_SET_ALL(cli->capabilities, CAP_NT_SMBS))
	{
		int f;
		f = cli_nt_create(cli, &(pipe_name[5]));
		if (f == -1)
		{
			fstring errstr;
			cli_safe_errstr(cli, errstr, sizeof(errstr) - 1);
			DEBUG(0,
			      ("cli_nt_session_open: cli_nt_create failed on pipe %s to machine %s.  Error was %s\n",
			       &(pipe_name[5]), cli->desthost, errstr));
			return False;
		}
		*fnum = (uint16)f;
	}
	else
	{
		int f;
		f = cli_open(cli, pipe_name, O_CREAT | O_RDWR, DENY_NONE);
		if (f == -1)
		{
			fstring errstr;
			cli_safe_errstr(cli, errstr, sizeof(errstr) - 1);
			DEBUG(0,
			      ("cli_nt_session_open: cli_open failed on pipe %s to machine %s.  Error was %s\n",
			       pipe_name, cli->desthost, errstr));
			return False;
		}
		*fnum = (uint16)f;
		/**************** Set Named Pipe State ***************/
		if (!rpc_pipe_set_hnd_state(cli, *fnum, pipe_name, 0x4300))
		{
			fstring errstr;
			cli_safe_errstr(cli, errstr, sizeof(errstr) - 1);
			DEBUG(0,
			      ("cli_nt_session_open: pipe hnd state failed.  Error was %s\n",
			       errstr)); cli_close(cli, *fnum);
			return False;
		}

	}

	return True;
}

/****************************************************************************
close the session
****************************************************************************/

void cli_nt_session_close(struct cli_state *cli, uint16 fnum)
{
	if (fnum != 0xffff)
	{
		cli_close(cli, fnum);
	}
}


/* create new connection.  strictly speaking, one arg should be
 * full dce/rpc format: e.g "ncacn_np:\\server\pipe\pipename" */
static cli_rpc_info *ncacn_np_connect_add(const char *pipe_name,
			     uint16 vuid,
			     const char *srv_name,
			     const struct ntuser_creds *ntc,
			     BOOL reuse, BOOL *is_new_connection)
{
	return ncacn_np_use_add(pipe_name, vuid, srv_name,
				ntc, reuse, is_new_connection);

}

/****************************************************************************
terminate client connection
****************************************************************************/
static void ncacn_np_connection_free(cli_rpc_info *con_info)
{
	struct ncacn_np *msrpc = (struct ncacn_np*)con_info;
	BOOL closed = False;
	DEBUG(10, ("msrpc smb connection\n"));
	ncacn_np_use_del(msrpc->smb->desthost,
			 msrpc->pipe_name,
			 msrpc->smb->nt.vuid, False, &closed);
}

/****************************************************************************
 get nt creds (HACK ALERT!) associated with an msrpc session.
****************************************************************************/
static struct ntdom_info *ncacn_np_conn_get_ntinfo(cli_rpc_info *con_info)
{
	struct ncacn_np *msrpc = (struct ncacn_np*)con_info;
	return &msrpc->smb->nt;
}


/****************************************************************************
get a server name associated with a connection associated with a
policy handle.
****************************************************************************/
static const char *ncacn_np_con_get_srvname(cli_rpc_info *con_info)
{
	struct ncacn_np *msrpc = (struct ncacn_np*)con_info;
	return msrpc->smb->desthost;
}

/****************************************************************************
 write to a pipe
****************************************************************************/
static BOOL ncacn_np_api_write(cli_rpc_info *con_info, prs_struct *data)
{
	struct ncacn_np *msrpc = (struct ncacn_np*)con_info;
	struct cli_state *cli = msrpc->smb;
	int fnum = msrpc->fnum;
	return cli_write(cli, fnum, 0x0008,
			 data->data, 0,
			 prs_data_size(data), prs_data_size(data)) > 0;
}

static BOOL ncacn_np_api_rcv_pdu(struct cli_connection *con,
		cli_rpc_info *con_info, prs_struct *rdata)
{
	struct ncacn_np *msrpc = (struct ncacn_np*)con_info;
	struct cli_state *cli = msrpc->smb;
	int fnum = msrpc->fnum;
	return cli_rcv_pdu(con, cli, fnum, rdata);
}

/* this allows us to detect dead servers. The cli->fd is set to -1 when
   we get an error */
static BOOL ncacn_np_con_ok(cli_rpc_info *con_info)
{
	struct ncacn_np *msrpc = (struct ncacn_np*)con_info;
	struct cli_state *cli;
	if (msrpc == NULL)
		return False;
	cli = msrpc->smb;
	if (cli->fd == -1)
		return False;
	return True;
}


static BOOL ncacn_np_api_send_rcv_pdu(struct cli_connection *con,
		cli_rpc_info *con_info, prs_struct *data,
			  prs_struct *rdata)
{
	struct ncacn_np *msrpc = (struct ncacn_np*)con_info;
	struct ntdom_info *nt = ncacn_np_conn_get_ntinfo(con_info);
	struct cli_state *cli = msrpc->smb;
	int fnum = msrpc->fnum;
	if (cli->fd == -1)
		return False;
	return cli_send_and_rcv_pdu(con, cli, fnum, data,
				    rdata, nt->max_xmit_frag);
}

static cli_connect_fns ncacn_np_fns = 
{
	ncacn_np_connect_add,
	ncacn_np_connection_free,
	ncacn_np_conn_get_ntinfo,
	ncacn_np_con_get_srvname,
	ncacn_np_api_write,
	ncacn_np_api_rcv_pdu,
	ncacn_np_con_ok,
	ncacn_np_api_send_rcv_pdu

};

cli_connect_fns *ncacn_np_get_fns(void)
{
	return &ncacn_np_fns;
}

