/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2001
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "includes.h"
#include "debug.h"
#include "rpc_parse.h"
#include "rpc_client.h"


/****************************************************************************
 send a request on an rpc pipe.
 ****************************************************************************/
static BOOL create_noauth_pdu(struct cli_connection *con,
				uint8 op_num,
				prs_struct *data, int data_start, int *data_end,
				prs_struct *dataa,
				uint8 *flags)
{
	/* fudge this, at the moment: create the header; memcpy the data.  oops. */
	prs_struct data_t;
	prs_struct hdr;
	int data_len;
	int frag_len;
	int max_recv_frag;
	char *d = prs_data(data, data_start);
	struct ntdom_info *nt = cli_conn_get_ntinfo(con);

	if (nt == NULL)
	{
		DEBUG(1, ("WARNING: create_noauth_pdu: no ntinfo for con\n"));
		return False;
	}

	max_recv_frag = nt->max_recv_frag;

	if (max_recv_frag == 0)
	{
		max_recv_frag = 256;
		DEBUG(1, ("WARNING: create_noauth_pdu: max_recv_frag==0, "
			  "setting to 0x%x as work-around\n", max_recv_frag));
	}

	*flags = 0;

	data_len = data->offset - data_start;

	DEBUG(20, ("create_noauth_pdu: data_start=0x%x, offset=0x%x, "
		   "data_len=0x%x, data_end=0x%x ",
		   data_start, data->offset, data_len, *data_end));

	if (data_start == 0)
	{
		(*flags) |= RPC_FLG_FIRST;
	}

	if (data_len > max_recv_frag)
	{
		data_len = max_recv_frag;
	}
	else
	{
		(*flags) |= RPC_FLG_LAST;
	}

	(*data_end) += data_len;

	/* happen to know that NTLMSSP authentication verifier is 16 bytes */
	frag_len = data_len + 0x18;

	DEBUGADD(20, ("-> data_len=0x%x, frag_len=0x%x, data_end=0x%x\n",
		      data_len, frag_len, *data_end));

	prs_init(&data_t   , 0       , 4, False);
	prs_init(&hdr      , frag_len, 4, False);

	prs_append_data(&data_t, d, data_len);
	data_t.end = data_t.data_size;
	data_t.offset = data_t.data_size;

	/* 1 context_id (causing problems) */
	create_rpc_request(&hdr, 1, op_num, (*flags), frag_len, 0);

	prs_init(dataa, 0, 4, False);
	prs_append_prs(dataa, &hdr);
	prs_append_prs(dataa, &data_t);

	prs_free_data(&hdr      );
	prs_free_data(&data_t   );

	DEBUG(100,("frag_len: 0x%x data_len: 0x%x data_calc_len: 0x%x\n",
		frag_len, data_len, prs_buf_len(&data_t)));

	if (frag_len != prs_buf_len(dataa))
	{
		DEBUG(0,("expected fragment length does not match\n"));

		return False;
	}

	return True;
}

/*******************************************************************
 creates a DCE/RPC bind request

 - initialises the parse structure.
 - dynamically allocates the header data structure
 - caller is expected to free the header data structure once used.

 ********************************************************************/
static BOOL create_rpc_noauth_bind_req(struct cli_connection *con,
				prs_struct *data,
				uint32 rpc_call_id,
                                RPC_IFACE *abstract, RPC_IFACE *transfer)
{
	prs_struct rhdr;
	prs_struct rhdr_rb;

	RPC_HDR_RB           hdr_rb;
	RPC_HDR              hdr;

	DEBUG(10,("create_rpc_noauth_bind_req\n"));

	prs_init(&rhdr     , 0x0, 4, False);
	prs_init(&rhdr_rb  , 0x0, 4, False);

	/* 0x1 context_id + create the bind request RPC_HDR_RB */
	make_rpc_hdr_rb(&hdr_rb, 0x1630, 0x1630, 0x0,
	                0x1, 0x1, 0x1, abstract, transfer);

	/* stream the bind request data */
	smb_io_rpc_hdr_rb("", &hdr_rb,  &rhdr_rb, 0);

	/* create the request RPC_HDR */
	make_rpc_hdr(&hdr, RPC_BIND, 0x0, rpc_call_id,
	             rhdr_rb.offset + 0x10, 0);

	smb_io_rpc_hdr("hdr"   , &hdr   , &rhdr, 0);

	if (rhdr.data == NULL || rhdr_rb.data == NULL) return False;

	prs_init(data, 0, 4, False);
	prs_append_prs(data, &rhdr);
	prs_append_prs(data, &rhdr_rb);

	prs_free_data(&rhdr     );
	prs_free_data(&rhdr_rb  );

	return True;
}

cli_auth_fns cli_noauth_fns = 
{
	create_rpc_noauth_bind_req,
	NULL,
	NULL,
	create_noauth_pdu,
	NULL
};
