/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2001
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "includes.h"
#include "libsamba.h"
#include "byteorder.h"
#include "libsmb.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "md5.h"
#include "hmacmd5.h"

#define CLI_SIGN "session key to client-to-server signing key magic constant"
#define CLI_SEAL "session key to client-to-server sealing key magic constant"
#define SRV_SIGN "session key to server-to-client signing key magic constant"
#define SRV_SEAL "session key to server-to-client sealing key magic constant"

extern pstring global_myname;

static void NTLMSSPcalc_ap( unsigned char *hash, unsigned char *data, int len)
{
    unsigned char index_i = hash[256];
    unsigned char index_j = hash[257];
    int ind;

    for (ind = 0; ind < len; ind++)
    {
        unsigned char tc;
        unsigned char t;

        index_i++;
        index_j += hash[index_i];

        tc = hash[index_i];
        hash[index_i] = hash[index_j];
        hash[index_j] = tc;

        t = hash[index_i] + hash[index_j];
        data[ind] = data[ind] ^ hash[t];
    }

    hash[256] = index_i;
    hash[257] = index_j;
}

/****************************************************************************
 decrypt data on an rpc pipe
 ****************************************************************************/
static BOOL decode_ntlmssp_pdu(struct cli_connection *con,
				prs_struct *rdata,
				int len, int auth_len)
{
	RPC_AUTH_NTLMSSP_CHK  chk;
	RPC_AUTH_NTLMSSP_CHK2 chk2;
	uint32 crc32;

	int data_len = len - 0x18 - auth_len - 8;
	char *reply_data = prs_data(rdata, 0x18);

	BOOL auth_verify;
	BOOL auth_seal  ;
	BOOL ntlmv2     ;

	ntlmssp_auth_struct *a;
	a = (ntlmssp_auth_struct *)cli_conn_get_auth_info(con);

	if (a == NULL)
	{
		return False;
	}

	auth_verify = IS_BITS_SET_ALL(a->ntlmssp_chal.neg_flags,
	                              NTLMSSP_NEGOTIATE_SIGN);
	auth_seal   = IS_BITS_SET_ALL(a->ntlmssp_chal.neg_flags,
	                              NTLMSSP_NEGOTIATE_SEAL);
	ntlmv2      = IS_BITS_SET_ALL(a->ntlmssp_chal.neg_flags,
	                              NTLMSSP_NEGOTIATE_NTLM2);

	DEBUG(5, ("decode_ntlmssp_pdu: len: %d auth_len: %d data_len: %d "
		  "verify=%s seal=%s ntlmv2=%s\n",
	          len, auth_len, data_len,
	          BOOLSTR(auth_verify), BOOLSTR(auth_seal), BOOLSTR(ntlmv2)));

	if (reply_data == NULL) return False;

	if (auth_seal)
	{
		if (ntlmv2)
		{
			DEBUG(10,("decode_ntlmssp_pdu: seal ntlmv2\n"));
		}
		else
		{
			DEBUG(10,("decode_ntlmssp_pdu: seal ntlmv1\n"));
		}

		dump_data(100, reply_data, data_len);

		if (ntlmv2)
		{
			NTLMSSPcalc_ap(a->srv_seal_hash, (uchar*)reply_data, data_len);
		}
		else
		{
			NTLMSSPcalc_ap(a->ntlmssp_hash, (uchar*)reply_data, data_len);
		}
		dump_data(100, reply_data, data_len);
	}

	if (auth_verify || auth_seal)
	{
		RPC_HDR_AUTH         rhdr_auth; 
		prs_struct auth_req;
		prs_init(&auth_req , 0x0, 4, True);
		prs_append_data(&auth_req,
		                prs_data(rdata, len - auth_len - 8),
		                8);
		smb_io_rpc_hdr_auth("hdr_auth", &rhdr_auth, &auth_req, 0);
		prs_free_data(&auth_req);

		if (!rpc_hdr_ntlmssp_auth_chk(&rhdr_auth))
		{
			return False;
		}
	}

	if (auth_verify && !ntlmv2)
	{
		prs_struct auth_verf;
		char *data = prs_data(rdata, len - auth_len);
		if (data == NULL) return False;

		DEBUG(10,("decode_ntlmssp_pdu: verify ntlmv1\n"));
		dump_data(100, data, auth_len);
		NTLMSSPcalc_ap(a->ntlmssp_hash, (uchar*)(data+4), auth_len - 4);
		prs_init(&auth_verf, 0x0, 4, True);
		prs_append_data(&auth_verf, data, 16);
		smb_io_rpc_auth_ntlmssp_chk("auth_sign", &chk, &auth_verf, 0);
		dump_data(100, data, auth_len);
		prs_free_data(&auth_verf);

		crc32 = crc32_calc_buffer(data_len, prs_data(rdata, 0x18));
		if (!rpc_auth_ntlmssp_chk(&chk, crc32 , a->ntlmssp_seq_num))
		{
			return False;
		}
		
		a->ntlmssp_seq_num++;
	}
	if (auth_verify && ntlmv2)
	{
		prs_struct auth_verf;
		char *data = prs_data(rdata, len - auth_len);
		char sum[16];
		char seq_num[4];

		if (data == NULL) return False;

		SIVAL(seq_num, 0, &a->ntlmssp_seq_num);

		DEBUG(10,("decode_ntlmssp_pdu: verify ntlmv2\n"));
		dump_data(100, data, auth_len);
		NTLMSSPcalc_ap(a->srv_sign_hash, (uchar*)(data+4), auth_len - 4);
		{
			prs_init(&auth_verf, 0x0, 4, True);
			prs_append_data(&auth_verf, data, 16);
			smb_io_rpc_auth_ntlmssp_chk2("auth_sign", &chk2, &auth_verf, 0);
			dump_data(100, data, auth_len);
			prs_free_data(&auth_verf);
		}

		hmac_md5_2d(a->srv_sign_const,
				seq_num, 4,
				prs_data(rdata, 0x18), data_len, 
				sum);

		if (!rpc_auth_ntlmssp_chk2(&chk2, sum , a->ntlmssp_seq_num))
		{
				return False;
		}

		a->ntlmssp_seq_num++;
	}
	return True;
}

/****************************************************************************
 send a request on an rpc pipe.
 ****************************************************************************/
static BOOL create_ntlmssp_pdu(struct cli_connection *con,
				uint8 op_num,
				prs_struct *data, int data_start, int *data_end,
				prs_struct *dataa,
				uint8 *flags)
{
	prs_struct data_t;
	prs_struct hdr;
	prs_struct hdr_auth;
	prs_struct auth_verf;
	int data_len;
	int frag_len;
	int auth_len;
	BOOL auth_verify;
	BOOL auth_seal;
	BOOL ntlmv2;
	uint32 crc32 = 0;
	char sum[16];

	struct ntdom_info *nt = cli_conn_get_ntinfo(con);
	ntlmssp_auth_struct *a;
	a = (ntlmssp_auth_struct *)cli_conn_get_auth_info(con);

	if (a == NULL)
	{
		return False;
	}

	(*flags) = 0;

	auth_verify = IS_BITS_SET_ALL(a->ntlmssp_chal.neg_flags,
	                              NTLMSSP_NEGOTIATE_SIGN);
	auth_seal   = IS_BITS_SET_ALL(a->ntlmssp_chal.neg_flags,
	                              NTLMSSP_NEGOTIATE_SEAL);
	ntlmv2      = IS_BITS_SET_ALL(a->ntlmssp_chal.neg_flags,
	                              NTLMSSP_NEGOTIATE_NTLM2);

	auth_len = (auth_verify ? 16 : 0);
	data_len = data->offset - data_start;

	if (data_start == 0)
	{
		(*flags) |= RPC_FLG_FIRST;
	}

	if (data_len > nt->max_recv_frag)
	{
		data_len = nt->max_recv_frag - (auth_len + (auth_verify ? 8 : 0) + 0x18);
	}
	else
	{
		(*flags) |= RPC_FLG_LAST;
	}

	(*data_end) += data_len;

	/* happen to know that NTLMSSP authentication verifier is 16 bytes */
	frag_len = data_len + auth_len + (auth_verify ? 8 : 0) + 0x18;

	prs_init(&data_t   , 0       , 4, False);
	prs_init(&hdr      , frag_len, 4, False);
	prs_init(&hdr_auth , 0       , 4, False);
	prs_init(&auth_verf, auth_len, 4, False);

	prs_append_data(&data_t, prs_data(data, data_start), data_len);
	data_t.end = data_t.data_size;
	data_t.offset = data_t.data_size;

        /* 0x0 context_id because it's causing problems... */
	create_rpc_request(&hdr, 0x1, op_num, (*flags),
	                                      frag_len, auth_len);

	if (auth_seal && !ntlmv2)
	{
		char *buf = prs_data(&data_t, 0);
		size_t len = prs_buf_len(&data_t);
		crc32 = crc32_calc_buffer(len, buf);
		NTLMSSPcalc_ap(a->ntlmssp_hash, (uchar*)buf, len);
		dump_data(10, buf, len);
	}

	if (auth_seal && ntlmv2)
	{
		char *buf = prs_data(&data_t, 0);
		size_t len = prs_buf_len(&data_t);
		char seq_num[4];

		SIVAL(seq_num, 0, &a->ntlmssp_seq_num);

		hmac_md5_2d(a->cli_sign_const,
				seq_num, 4,
				buf, len,
				sum);

		NTLMSSPcalc_ap(a->cli_seal_hash, (uchar*)buf, len);
		dump_data(10, buf, len);
	}

	if (auth_seal || auth_verify)
	{
		RPC_HDR_AUTH         rhdr_auth;

		make_rpc_hdr_auth(&rhdr_auth, 0x0a, 0x06, 0x0, (auth_verify ? 0x1 : 0));
		smb_io_rpc_hdr_auth("hdr_auth", &rhdr_auth, &hdr_auth, 0);
	}

	if (auth_verify && !ntlmv2)
	{
		RPC_AUTH_NTLMSSP_CHK chk;

		make_rpc_auth_ntlmssp_chk(&chk, NTLMSSP_SIGN_VERSION, crc32, a->ntlmssp_seq_num++);
		smb_io_rpc_auth_ntlmssp_chk("auth_sign", &chk, &auth_verf, 0);
		NTLMSSPcalc_ap(a->ntlmssp_hash, (uchar*)prs_data(&auth_verf, 4), 12);
	}

	if (auth_verify && ntlmv2)
	{
		RPC_AUTH_NTLMSSP_CHK2 chk2;

		make_rpc_auth_ntlmssp_chk2(&chk2, NTLMSSP_SIGN_VERSION, sum, a->ntlmssp_seq_num++);
		smb_io_rpc_auth_ntlmssp_chk2("auth_sign", &chk2, &auth_verf, 0);
		NTLMSSPcalc_ap(a->cli_seal_hash, (uchar*)prs_data(&auth_verf, 4), 12);
	}

	prs_init(dataa, 0, 4, False);

	prs_append_prs(dataa, &hdr);
	prs_append_prs(dataa, &data_t);

	if (auth_seal || auth_verify)
	{
		prs_append_prs(dataa, &hdr_auth);
		prs_append_prs(dataa, &auth_verf);
	}

	prs_free_data(&hdr_auth );
	prs_free_data(&auth_verf);
	prs_free_data(&hdr      );
	prs_free_data(&data_t   );

	DEBUG(100,("frag_len: 0x%x data_len: 0x%x data_calc_len: 0x%x\n",
		frag_len, data_len, prs_buf_len(&data_t)));

	if (frag_len != prs_buf_len(dataa))
	{
		DEBUG(0,("expected fragment length does not match\n"));

		return False;
	}

	return True;
}

/*******************************************************************
 creates a DCE/RPC bind request

 - initialises the parse structure.
 - dynamically allocates the header data structure
 - caller is expected to free the header data structure once used.

 ********************************************************************/
static BOOL create_ntlmssp_bind_req(struct cli_connection *con,
				prs_struct *data,
				uint32 rpc_call_id,
                                RPC_IFACE *abstract, RPC_IFACE *transfer)
{
	prs_struct rhdr;
	prs_struct rhdr_rb;
	prs_struct rhdr_auth;
	prs_struct auth_req;

	RPC_HDR_RB           hdr_rb;
	RPC_HDR              hdr;
	RPC_HDR_AUTH         hdr_auth;
	RPC_AUTH_VERIFIER auth_verifier;
	RPC_AUTH_NTLMSSP_NEG ntlmssp_neg;

	struct ntuser_creds *usr;
	usr = (struct ntuser_creds*)cli_conn_get_auth_creds(con);

	if (usr == NULL)
	{
		DEBUG(10,("create_ntlmssp_bind_req: NULL user creds\n"));
		return False;
	}

	prs_init(&rhdr     , 0x0, 4, False);
	prs_init(&rhdr_rb  , 0x0, 4, False);
	prs_init(&rhdr_auth, 0x0, 4, False);
	prs_init(&auth_req , 0x0, 4, False);

	/* 0x1 context + create the bind request RPC_HDR_RB - zero oontext_id */
	make_rpc_hdr_rb(&hdr_rb, 0x1630, 0x1630, 0x0,
	                0x1, 0x1, 0x1, abstract, transfer);

	/* stream the bind request data */
	smb_io_rpc_hdr_rb("", &hdr_rb,  &rhdr_rb, 0);

	make_rpc_hdr_auth(&hdr_auth, 0x0a, 0x06, 0x08, 0x1);
	smb_io_rpc_hdr_auth("hdr_auth", &hdr_auth, &rhdr_auth, 0);

	make_rpc_auth_verifier(&auth_verifier,
			       "NTLMSSP", NTLMSSP_NEGOTIATE);

	smb_io_rpc_auth_verifier("auth_verifier", &auth_verifier, &auth_req, 0);

	make_rpc_auth_ntlmssp_neg(&ntlmssp_neg,
			       usr->ntlmssp_flags, global_myname, usr->domain);

	smb_io_rpc_auth_ntlmssp_neg("ntlmssp_neg", &ntlmssp_neg, &auth_req, 0);

	/* create the request RPC_HDR */
	make_rpc_hdr(&hdr, RPC_BIND, 0x0, rpc_call_id,
	             auth_req .offset + rhdr_auth.offset +
	             rhdr_rb.offset + 0x10,
	             auth_req .offset);

	smb_io_rpc_hdr("hdr"   , &hdr   , &rhdr, 0);

	if (rhdr.data == NULL || rhdr_rb.data == NULL) return False;

	prs_init(data, 0, 4, False);
	prs_append_prs(data, &rhdr);
	prs_append_prs(data, &rhdr_rb);
	prs_append_prs(data, &rhdr_auth);
	prs_append_prs(data, &auth_req);

	prs_free_data(&rhdr     );
	prs_free_data(&rhdr_rb  );
	prs_free_data(&rhdr_auth);
	prs_free_data(&auth_req );

	return cli_conn_set_auth_info(con,
	             (void*)malloc(sizeof(struct ntlmssp_auth_struct)));
}

static BOOL decode_ntlmssp_bind_resp(struct cli_connection *con,
				prs_struct *rdata)
{
	BOOL valid_ack = True;

	ntlmssp_auth_struct *a;
	a = (ntlmssp_auth_struct *)cli_conn_get_auth_info(con);

	if (a == NULL)
	{
		return False;
	}

	if (valid_ack)
	{
		RPC_HDR_AUTH rhdr_auth;
		smb_io_rpc_hdr_auth("", &rhdr_auth, rdata, 0);
		if (rdata->offset == 0 ||
		    !rpc_hdr_ntlmssp_auth_chk(&rhdr_auth))
		{
			valid_ack = False;
		}
	}
	if (valid_ack)
	{
		RPC_AUTH_VERIFIER rhdr_verf;
		smb_io_rpc_auth_verifier("", &rhdr_verf, rdata, 0);
		if (rdata->offset == 0 ||
		    !rpc_auth_verifier_chk(&rhdr_verf,
		                                   "NTLMSSP",
		                                    NTLMSSP_CHALLENGE))
		{
			valid_ack = False;
		}
	}
	if (valid_ack)
	{
		smb_io_rpc_auth_ntlmssp_chal("", &a->ntlmssp_chal, rdata, 0);
		if (rdata->offset == 0) valid_ack = False;
	}
	return valid_ack;
}

/*******************************************************************
 creates a DCE/RPC bind authentication response

 - initialises the parse structure.
 - dynamically allocates the header data structure
 - caller is expected to free the header data structure once used.

 ********************************************************************/
static BOOL create_ntlmssp_rpc_bind_resp(struct pwd_info *pwd,
				char *domain, char *user_name, char *my_name,
				uint32 ntlmssp_cli_flgs,
				uint32 rpc_call_id,
				prs_struct *rhdr)
{
	RPC_HDR           hdr;
	RPC_HDR_AUTHA     hdr_autha;
	RPC_AUTH_VERIFIER auth_verifier;
	uchar lm_owf[24];
	uchar nt_owf[128];
	size_t nt_owf_len;
	prs_struct rhdr_autha;
	prs_struct auth_resp;

	prs_init(&rhdr_autha, 0, 4, MARSHALL);
	prs_init(&auth_resp, 0, 4, MARSHALL);

        /* key_id 0x1 is context_id??? */
	make_rpc_hdr_autha(&hdr_autha, 0x1630, 0x1630, 0x0a, 0x06, 0x00,
                                0x1);
	smb_io_rpc_hdr_autha("hdr_autha", &hdr_autha, &rhdr_autha, 0);
	prs_realloc_data(&rhdr_autha, rhdr_autha.offset);

	make_rpc_auth_verifier(&auth_verifier,
			       "NTLMSSP", NTLMSSP_AUTH);

	smb_io_rpc_auth_verifier("auth_verifier", &auth_verifier, &auth_resp, 0);
	prs_realloc_data(&auth_resp, auth_resp.offset);

	pwd_get_lm_nt_owf(pwd, lm_owf, nt_owf, &nt_owf_len);

	create_ntlmssp_resp(lm_owf, nt_owf, nt_owf_len,
			domain, user_name, my_name, ntlmssp_cli_flgs,
                                &auth_resp);

	/* create the request RPC_HDR */
	make_rpc_hdr(&hdr, RPC_BINDRESP, 0x0, rpc_call_id,
	             auth_resp.offset + rhdr_autha.offset + 0x10,
	             auth_resp.offset);

	smb_io_rpc_hdr("hdr"   , &hdr   , rhdr, 0);
	prs_realloc_data(rhdr, rhdr->offset);

	if (rhdr->data == NULL || rhdr_autha.data == NULL) return False;

	/***/
	/*** link rpc header and authentication responses ***/
	/***/

	prs_append_prs(rhdr, &rhdr_autha);
	prs_append_prs(rhdr, &auth_resp);

	prs_free_data(&rhdr_autha);
	prs_free_data(&auth_resp);

	return True;
}

static void calc_hash(unsigned char *hash, char *k2, int k2l)
{
	unsigned char j = 0;
	int ind;

	for (ind = 0; ind < 256; ind++)
	{
		hash[ind] = (unsigned char)ind;
	}

	for (ind = 0; ind < 256; ind++)
	{
		unsigned char tc;

		j += (hash[ind] + k2[ind%k2l]);

		tc = hash[ind];
		hash[ind] = hash[j];
		hash[j] = tc;
	}

	hash[256] = 0;
	hash[257] = 0;
}

static void calc_ntlmv2_hash(unsigned char *hash,
		char *digest,
		const char *p24, const char *constant)
{
	struct MD5Context ctx3;

	MD5Init(&ctx3);
	MD5Update(&ctx3, p24, 5);
	MD5Update(&ctx3, constant, strlen(CLI_SIGN));
	MD5Final(digest, &ctx3);

	calc_hash(hash, digest, 16);
}

/*******************************************************************
 creates a DCE/RPC bind continue request

 - initialises the parse structure.
 - dynamically allocates the header data structure
 - caller is expected to free the header data structure once used.

 ********************************************************************/
static BOOL create_ntlmssp_bind_cont(struct cli_connection *con,
				prs_struct *dataa,
				uint32 rpc_call_id)
{
	BOOL ret = False;

	BOOL ntlmv2     ;
	unsigned char p24[24];
	unsigned char lm_owf[24];
	unsigned char lm_hash[16];
	unsigned char usr_sess_key[16];

	prs_struct hdra;

	struct ntuser_creds *usr;
	ntlmssp_auth_struct *a;
	a = (ntlmssp_auth_struct *)cli_conn_get_auth_info(con);
	usr = (struct ntuser_creds*)cli_conn_get_auth_creds(con);

	DEBUG(5,("Bind RPC Cont\n"));

	if (a == NULL)
	{
		return False;
	}

	prs_init(&hdra     , 0x0, 4, False);

	pwd_make_lm_nt_owf(&usr->pwd, a->ntlmssp_chal.challenge, usr_sess_key);

	create_ntlmssp_rpc_bind_resp(&usr->pwd, usr->domain,
			     usr->user_name, global_myname,
			     a->ntlmssp_chal.neg_flags,
			     rpc_call_id,
			     &hdra);
			    
	ntlmv2      = IS_BITS_SET_ALL(a->ntlmssp_chal.neg_flags,
	                              NTLMSSP_NEGOTIATE_NTLM2);

	cli_set_con_usr_sesskey(con, usr_sess_key);
	pwd_get_lm_nt_owf(&usr->pwd, lm_owf, NULL, NULL);
	pwd_get_lm_nt_16(&usr->pwd, lm_hash, NULL);
	NTLMSSPOWFencrypt(lm_hash, lm_owf, p24);

	if (ntlmv2)
	{
		calc_ntlmv2_hash(a->cli_sign_hash, a->cli_sign_const, p24, CLI_SIGN);
		calc_ntlmv2_hash(a->cli_seal_hash, a->cli_seal_const, p24, CLI_SEAL);
		calc_ntlmv2_hash(a->srv_sign_hash, a->srv_sign_const, p24, SRV_SIGN);
		calc_ntlmv2_hash(a->srv_seal_hash, a->srv_seal_const, p24, SRV_SEAL);
	}
	else
	{
		char k2[8];
		memcpy(k2, p24, 5);
		k2[5] = 0xe5;
		k2[6] = 0x38;
		k2[7] = 0xb0;

		calc_hash(a->ntlmssp_hash, k2, 8);
	}
	a->ntlmssp_seq_num = 0;

	ZERO_STRUCT(lm_hash);

	prs_init(dataa, 0, 4, False);
	ret = prs_copy(dataa, &hdra);

	prs_free_data(&hdra);

	return ret;
}

cli_auth_fns cli_ntlmssp_fns =
{
	create_ntlmssp_bind_req,
	decode_ntlmssp_bind_resp,
	create_ntlmssp_bind_cont,
	create_ntlmssp_pdu,
	decode_ntlmssp_pdu
};
