/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-2000,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2000,
 *  Copyright (C) Paul Ashton                  1997-2000.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "includes.h"
#include "libsamba.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "nterr.h"
#include "doserr.h"
#include "others.h"


/****************************************************************************
do a REG Open Policy
****************************************************************************/
static uint32 reg_open_hkXX_int(struct cli_connection *con,
				uint16 op_code, const char *op_name,
				uint16 unknown_0, uint32 access_mask,
				POLICY_HND *hnd)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_OPEN_HKXX q_o;
	REG_R_OPEN_HKXX r_o;
	uint32 status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return NT_STATUS_UNSUCCESSFUL;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Open (%s)\n", op_name));

	make_reg_q_open_hkXX(&q_o, unknown_0, access_mask);

	/* turn parameters into data stream */
	if (reg_io_q_open_hkXX("", &q_o, &buf, 0)
	    && rpc_con_pipe_req(con, op_code, &buf, &rbuf)
	    && reg_io_r_open_hkXX("", &r_o, &rbuf, 0))
	{
		status = r_o.status;

		if (status != 0)
		{
			/* report error code */
			DEBUG(1, ("%s: %s\n", op_name,
				  get_nt_error_msg(status)));
		}

		*hnd = r_o.pol;
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return status;
}

/****************************************************************************
do a REG Open Policy
****************************************************************************/
BOOL reg_connect(const char *srv_name,
				const char *full_keyname,
				char *key_name,
				uint32 access_mask,
				POLICY_HND *reg_hnd)
{
	BOOL res = True;
	uint32 reg_type = 0;

	struct cli_connection *con = NULL;

	if (!cli_connection_init(srv_name, PIPE_WINREG, &con))
	{
		return False;
	}

	ZERO_STRUCTP(reg_hnd);

	if (full_keyname == NULL)
	{
		return False;
	}

	/*
	 * open registry receive a policy handle
	 */

	if (!reg_split_key(full_keyname, &reg_type, key_name))
	{
		DEBUG(0,("reg_connect: unrecognised key name %s\n",
		          full_keyname));	
		return False;
	}

	switch (reg_type)
	{
		case HKEY_CLASSES_ROOT:
		{
			res = res ? reg_open_hkXX_int(con,
					REG_OPEN_HKCR, "REG_OPEN_HKCR",
					0x5428, access_mask,
					reg_hnd) == 0x0 : False;
			break;
		}
	
		case HKEY_LOCAL_MACHINE:
		{
			res = res ? reg_open_hkXX_int(con,
					REG_OPEN_HKLM, "REG_OPEN_HKLM",
					0x84E0, access_mask,
					reg_hnd) == 0x0 : False;
			break;
		}
	
		case HKEY_USERS:
		{
			res = res ? reg_open_hkXX_int(con,
					REG_OPEN_HKU, "REG_OPEN_HKU",
					0x84E0, access_mask,
					reg_hnd) == 0x0 : False;
			break;
		}

		case HKEY_PERFORMANCE_DATA:
		{
			res = res ? reg_open_hkXX_int(con,
					REG_OPEN_HKPD, "REG_OPEN_HKPD",
					0x86e8, access_mask,
					reg_hnd) == 0x0 : False;
			break;
		}

		default:
		{
			DEBUG(0,("reg_connect: unrecognised hive key\n"));	
			return False;
		}
	}

	if (res)
	{
		if (!register_policy_hnd(get_global_hnd_cache(),
					 cli_con_sec_ctx(con),
		                         reg_hnd, access_mask)
		    || !set_policy_con(get_global_hnd_cache(), reg_hnd, con, 
				       cli_connection_unlink))
		{
			cli_connection_unlink(con);
			return False;
		}
	}

	return res;
}

/****************************************************************************
do a REG Unknown 0xB command.  sent after a create key or create value.
this might be some sort of "sync" or "refresh" command, sent after
modification of the registry...
****************************************************************************/
BOOL reg_flush_key(POLICY_HND *hnd)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_FLUSH_KEY q_o;
	REG_R_FLUSH_KEY r_o;
	BOOL valid_query = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Unknown 0xB\n"));

	make_reg_q_flush_key(&q_o, hnd);

	/* turn parameters into data stream */
	if (reg_io_q_flush_key("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_FLUSH_KEY, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		reg_io_r_flush_key("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_FLUSH_KEY: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			valid_query = True;
		}
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_query;
}

/****************************************************************************
do a REG Query Key
****************************************************************************/
BOOL reg_query_key(POLICY_HND *hnd,
				char *key_class, uint32 *class_len,
				uint32 *num_subkeys, uint32 *max_subkeylen,
				uint32 *max_subkeysize, uint32 *num_values,
				uint32 *max_valnamelen, uint32 *max_valbufsize,
				uint32 *sec_desc, NTTIME *mod_time)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_QUERY_KEY q_o;
	REG_R_QUERY_KEY r_o;
	BOOL valid_query = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Query Key\n"));

	make_reg_q_query_key(&q_o, hnd, *class_len);

	/* turn parameters into data stream */
	if (reg_io_q_query_key("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_QUERY_KEY, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		reg_io_r_query_key("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_QUERY_KEY: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			valid_query = True;
			
			*class_len      = r_o.hdr_class.uni_max_len;
			unistr2_to_ascii(key_class, &r_o.uni_class, sizeof(fstring)-1);
			*num_subkeys    = r_o.num_subkeys   ;
			*max_subkeylen  = r_o.max_subkeylen ;
			*max_subkeysize = r_o.max_subkeysize;
			*num_values     = r_o.num_values    ;
			*max_valnamelen = r_o.max_valnamelen;
			*max_valbufsize = r_o.max_valbufsize;
			*sec_desc       = r_o.sec_desc      ;
			*mod_time       = r_o.mod_time      ;
		}
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_query;
}

/****************************************************************************
do a REG Unknown 1A
****************************************************************************/
BOOL reg_unknown_1a(POLICY_HND *hnd, uint32 *unk)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_UNK_1A q_o;
	REG_R_UNK_1A r_o;
	BOOL valid_query = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Unknown 1a\n"));

	make_reg_q_unk_1a(&q_o, hnd);

	/* turn parameters into data stream */
	if (reg_io_q_unk_1a("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_UNK_1A, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		reg_io_r_unk_1a("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_UNK_1A: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			valid_query = True;
			if (unk)
			{
				*unk = r_o.unknown;
			}
			if (r_o.unknown != 5)
			{
				DEBUG(4,("REG Unknown 1a response: %x\n", r_o.unknown));
			}
		}
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_query;
}

/****************************************************************************
do a REG Query Info
****************************************************************************/
BOOL reg_query_info(POLICY_HND *hnd, const char *val_name,
		    uint32 *type, BUFFER2 *buffer)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_INFO q_o;
	REG_R_INFO r_o;
	BOOL valid_query = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Query Info\n"));

	make_reg_q_info(&q_o, hnd, val_name, 4, 0, buffer->buf_max_len);

	/* turn parameters into data stream */
	if (reg_io_q_info("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_INFO, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		r_o.type = *type;
		r_o.uni_val = buffer;

		reg_io_r_info("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_INFO: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			valid_query = True;
		}
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_query;
}

/****************************************************************************
do a REG Set Key Security 
****************************************************************************/
BOOL reg_set_key_sec(POLICY_HND *hnd,
				uint32 sec_info,
				uint32 sec_buf_size, SEC_DESC *sec_buf)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_SET_KEY_SEC q_o;
	REG_R_SET_KEY_SEC r_o;
	BOOL valid_query = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Set Key security.\n"));

	make_reg_q_set_key_sec(&q_o, hnd, sec_info, sec_buf_size, sec_buf);

	/* turn parameters into data stream */
	if (reg_io_q_set_key_sec("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_SET_KEY_SEC, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		reg_io_r_set_key_sec("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			valid_query = True;
		}
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_query;
}


/****************************************************************************
do a REG Query Key Security 
****************************************************************************/
BOOL reg_get_key_sec(POLICY_HND *hnd,
				uint32 sec_info,
				uint32 *sec_buf_size, SEC_DESC_BUF *sec_buf)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_GET_KEY_SEC q_o;
	REG_R_GET_KEY_SEC r_o;
	BOOL valid_query = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG query key security.  buf_size: %d\n", *sec_buf_size));

	make_reg_q_get_key_sec(&q_o, hnd, sec_info, *sec_buf_size, sec_buf);

	/* turn parameters into data stream */
	if (reg_io_q_get_key_sec("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_GET_KEY_SEC, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		r_o.data = sec_buf;
		if (*sec_buf_size != 0)
		{
			sec_buf->sec = (SEC_DESC*)malloc(*sec_buf_size);
		}
		reg_io_r_get_key_sec("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status == 0x0000007a)
		{
			/*
			 * get the maximum buffer size: it was too small
			 */
			(*sec_buf_size) = r_o.hdr_sec.buf_max_len;
			DEBUG(5, ("sec_buf_size too small.  use %d\n",
				  *sec_buf_size));
			valid_query = True;
		}
		else if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_GET_KEY_SEC: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}
		else
		{
			valid_query = True;
			(*sec_buf_size) = r_o.data->len;
		}
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_query;
}

/****************************************************************************
do a REG Delete Value
****************************************************************************/
BOOL reg_delete_val(POLICY_HND *hnd, char *val_name)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_DELETE_VALUE q_o;
	REG_R_DELETE_VALUE r_o;
	BOOL valid_delete = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Delete Value: %s\n", val_name));

	make_reg_q_delete_val(&q_o, hnd, val_name);

	/* turn parameters into data stream */
	if (reg_io_q_delete_val("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_DELETE_VALUE, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		reg_io_r_delete_val("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_DELETE_VALUE: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			valid_delete = True;
		}
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_delete;
}

/****************************************************************************
do a REG Delete Key
****************************************************************************/
BOOL reg_delete_key(POLICY_HND *hnd, char *key_name)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_DELETE_KEY q_o;
	REG_R_DELETE_KEY r_o;
	BOOL valid_delete = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Delete Key: %s\n", key_name));

	make_reg_q_delete_key(&q_o, hnd, key_name);

	/* turn parameters into data stream */
	if (reg_io_q_delete_key("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_DELETE_KEY, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		reg_io_r_delete_key("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_DELETE_KEY: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			valid_delete = True;
		}
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_delete;
}

/****************************************************************************
do a REG Create Key
****************************************************************************/
BOOL reg_create_key(POLICY_HND *hnd,
				char *key_name, char *key_class,
				SEC_ACCESS *sam_access,
				POLICY_HND *key)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_CREATE_KEY q_o;
	REG_R_CREATE_KEY r_o;
	BOOL valid_create = False;
	SEC_DESC sec;
	SEC_DESC_BUF sec_buf;
	int sec_len;

	ZERO_STRUCT(sec);
	ZERO_STRUCT(sec_buf);
	ZERO_STRUCT(q_o);

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Create Key: %s %s 0x%08x\n", key_name, key_class,
		sam_access != NULL ? sam_access->mask : 0));

	sec_len = make_sec_desc(&sec, 1, SEC_DESC_SELF_RELATIVE,
	                        NULL, NULL, NULL, NULL);

	DEBUG(10,("make_sec_desc: len = %d\n", sec_len));

	make_reg_q_create_key(&q_o, hnd, key_name, key_class, sam_access,
	                      &sec_buf, sec_len, &sec);

	/* turn parameters into data stream */
	if (reg_io_q_create_key("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_CREATE_KEY, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		reg_io_r_create_key("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_CREATE_KEY: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			valid_create = True;
			*key = r_o.key_pol;
		}
	}

	free_sec_desc(&sec);

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_create;
}

/****************************************************************************
do a REG Enum Key
****************************************************************************/
BOOL reg_enum_key(POLICY_HND *hnd,
				int key_index, char *key_name,
				uint32 *unk_1, uint32 *unk_2,
				time_t *mod_time)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_ENUM_KEY q_o;
	REG_R_ENUM_KEY r_o;
	BOOL valid_query = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Enum Key\n"));

	make_reg_q_enum_key(&q_o, hnd, key_index);

	/* turn parameters into data stream */
	if (reg_io_q_enum_key("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_ENUM_KEY, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		reg_io_r_enum_key("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_ENUM_KEY: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			valid_query = True;
			(*unk_1) = r_o.unknown_1;
			(*unk_2) = r_o.unknown_2;
			unistr_to_ascii(key_name, r_o.key_name.str.buffer,
					sizeof(fstring)-1);
			(*mod_time) = nt_time_to_unix(&r_o.time);
		}
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_query;
}

/****************************************************************************
do a REG Create Value
****************************************************************************/
BOOL reg_create_val(POLICY_HND *hnd,
				char *val_name, uint32 type, BUFFER3 *data)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_CREATE_VALUE q_o;
	REG_R_CREATE_VALUE r_o;
	BOOL valid_create = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Create Value: %s\n", val_name));

	make_reg_q_create_val(&q_o, hnd, val_name, type, data);

	/* turn parameters into data stream */
	if (reg_io_q_create_val("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_CREATE_VALUE, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		reg_io_r_create_val("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_CREATE_VALUE: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			valid_create = True;
		}
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_create;
}

/****************************************************************************
do a REG Enum Value
****************************************************************************/
BOOL reg_enum_val(POLICY_HND *hnd,
				int val_index, int max_valnamelen, int max_valbufsize,
				fstring val_name,
				uint32 *val_type, BUFFER2 *value)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_ENUM_VALUE q_o;
	REG_R_ENUM_VALUE r_o;
	BOOL valid_query = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Enum Value\n"));

	make_reg_q_enum_val(&q_o, hnd, val_index, max_valnamelen, max_valbufsize);

	/* turn parameters into data stream */
	if (reg_io_q_enum_val("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_ENUM_VALUE, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);
		r_o.buf_value = value;

		reg_io_r_enum_val("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_ENUM_VALUE: %s\n",
				  get_nt_error_msg(r_o.status)));
			if (r_o.status != ERRmoredata)
				p = False;
		}

		if (p)
		{
			valid_query = True;
			(*val_type) = r_o.type;
			unistr2_to_ascii(val_name, &r_o.uni_name, sizeof(fstring)-1);
		}
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_query;
}

/****************************************************************************
do a REG Open Key
****************************************************************************/
BOOL reg_open_entry(POLICY_HND *hnd,
				char *key_name, uint32 unk_0,
				POLICY_HND *key_hnd)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_OPEN_ENTRY q_o;
	REG_R_OPEN_ENTRY r_o;
	BOOL valid_pol = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Open Entry\n"));

	make_reg_q_open_entry(&q_o, hnd, key_name, unk_0);

	/* turn parameters into data stream */
	if (reg_io_q_open_entry("", &q_o, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_OPEN_ENTRY, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		reg_io_r_open_entry("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_OPEN_ENTRY: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			struct cli_connection *con = NULL;

			if (!cli_connection_get(hnd, &con))
			{
				return False;
			}

			*key_hnd = r_o.pol;
			valid_pol = cli_pol_link(key_hnd, hnd);
		}
	}

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_pol;
}

/****************************************************************************
do a REG Close
****************************************************************************/
BOOL reg_close(POLICY_HND *hnd)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_CLOSE q_c;
	REG_R_CLOSE r_c;
	BOOL valid_close = False;

	if (hnd == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Close\n"));

	/* store the parameters */
	make_reg_q_close(&q_c, hnd);

	/* turn parameters into data stream */
	if (reg_io_q_close("", &q_c, &buf, 0) &&
	    rpc_hnd_pipe_req(hnd, REG_CLOSE, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_c);

		reg_io_r_close("", &r_c, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_c.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_CLOSE: %s\n",
				  get_nt_error_msg(r_c.status)));
			p = False;
		}

		if (p)
		{
			valid_close = True;
		}
	}

	close_policy_hnd(get_global_hnd_cache(), hnd);

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_close;
}

/****************************************************************************
do a REG Shutdown Server
****************************************************************************/
BOOL reg_shutdown(const char *srv_name,
			const char *msg, uint32 timeout, uint16 flags)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_SHUTDOWN q_o;
	REG_R_SHUTDOWN r_o;
	BOOL valid_shutdown = False;

	struct cli_connection *con = NULL;

	if (!cli_connection_init(srv_name, PIPE_WINREG, &con))
	{
		return False;
	}

	if (msg == NULL)
		return False;

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Shutdown: (timeout: %d secs) %s\n", timeout, msg));

	make_reg_q_shutdown(&q_o, msg, timeout, flags);

	/* turn parameters into data stream */
	if (reg_io_q_shutdown("", &q_o, &buf, 0) &&
	    rpc_con_pipe_req(con, REG_SHUTDOWN, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		reg_io_r_shutdown("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_SHUTDOWN: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			valid_shutdown = True;
		}
	}

	cli_connection_unlink(con);

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_shutdown;
}


/****************************************************************************
do a REG Abort Shutdown
****************************************************************************/
BOOL reg_abort_shutdown(const char *srv_name)
{
	prs_struct rbuf;
	prs_struct buf; 
	REG_Q_ABORT_SHUTDOWN q_o;
	REG_R_ABORT_SHUTDOWN r_o;
	BOOL valid_abort = False;

	struct cli_connection *con = NULL;

	if (!cli_connection_init(srv_name, PIPE_WINREG, &con))
	{
		return False;
	}

	prs_init(&buf , 0, 4, False);
	prs_init(&rbuf, 0, 4, True );

	DEBUG(4,("REG Abort Shutdown\n"));

	make_reg_q_abort_shutdown(&q_o);

	/* turn parameters into data stream */
	if (reg_io_q_abort_shutdown("", &q_o, &buf, 0) &&
	    rpc_con_pipe_req(con, REG_ABORT_SHUTDOWN, &buf, &rbuf))
	{
		BOOL p;

		ZERO_STRUCT(r_o);

		reg_io_r_abort_shutdown("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0, ("REG_ABORT_SHUTDOWN: %s\n",
				  get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			valid_abort = True;
		}
	}

	cli_connection_unlink(con);

	prs_free_data(&rbuf);
	prs_free_data(&buf );

	return valid_abort;
}
