/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Samba memory buffer functions
   Copyright (C) Andrew Tridgell              1992-2000
   Copyright (C) Luke Kenneth Casson Leighton 1996-2000
   Copyright (C) Elrond                            2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include <errno.h>
#include "libsamba.h"
#include "byteorder.h"
#include "libsmb.h"
#include "rpc_parse.h"
#include "sma.h"

#undef  DBGC_CLASS
#define DBGC_CLASS DBGC_RPC


#define DBG_RW_PCVAL(charmode,string,depth,base,read,inbuf,outbuf,len) \
	{ RW_PCVAL(read,inbuf,outbuf,len) \
	DEBUGADD(5+depth,("%s%04x %s: ", \
             tab_depth(depth), base,string)); \
    if (charmode) print_asc(depth+5, (unsigned char*)(outbuf), (len)); else \
	{ uint32 idx; for (idx = 0; idx < len; idx++) { DEBUGADD(5+depth,("%02x ", (outbuf)[idx])); } } \
	DEBUGADD(5+depth,("\n")); } 

#define DBG_RW_PSVAL(charmode,string,depth,base,read,big_endian,inbuf,outbuf,len) \
	{ RW_PSVAL(read,big_endian,inbuf,outbuf,len) \
	DEBUG(5+depth,("%s%04x %s: ", \
             tab_depth(depth), base,string)); \
    if (charmode) print_asc(depth+5, (unsigned char*)(outbuf), 2*(len)); else \
	{ uint32 idx; for (idx = 0; idx < len; idx++) { DEBUG(5+depth,("%04x ", (outbuf)[idx])); } } \
	DEBUG(5+depth,("\n")); }

#define DBG_RW_PIVAL(charmode,string,depth,base,read,big_endian,inbuf,outbuf,len) \
	{ RW_PIVAL(read,big_endian,inbuf,outbuf,len) \
	DEBUG(5+depth,("%s%04x %s: ", \
             tab_depth(depth), base,string)); \
    if (charmode) print_asc(depth+5, (unsigned char*)(outbuf), 4*(len)); else \
	{ uint32 idx; for (idx = 0; idx < len; idx++) { DEBUG(5+depth,("%08x ", (outbuf)[idx])); } } \
	DEBUG(5+depth,("\n")); }

#define DBG_RW_CVAL(string,depth,base,read,inbuf,outbuf) \
	{ RW_CVAL(read,inbuf,outbuf,0) \
	DEBUGADD(5+depth,("%s%04x %s: %02x\n", \
                 tab_depth(depth), base, string, outbuf)); }

#define DBG_RW_SVAL(string,depth,base,read,big_endian,inbuf,outbuf) \
	{ RW_SVAL(read,big_endian,inbuf,outbuf,0) \
	DEBUGADD(5+depth,("%s%04x %s: %04x\n", \
                 tab_depth(depth), base, string, outbuf)); }

#define DBG_RW_IVAL(string,depth,base,read,big_endian,inbuf,outbuf) \
	{ RW_IVAL(read,big_endian,inbuf,outbuf,0) \
	DEBUGADD(5+depth,("%s%04x %s: %08x\n", \
                 tab_depth(depth), base, string, outbuf)); }


#define PRS_ALIGN(ps, size) \
	if (ps->align_fn) \
	{ \
		if (!ps->align_fn(ps, size)) \
		{ \
			DEBUG(1, ("align for size %d failed\n", size)); \
			return False; \
		} \
	} 


static void prs_clean_pending(prs_struct *ps, BOOL show);


/*******************************************************************
 SMA stuff
 ********************************************************************/

static SMA_REGION *prs_sma_region = NULL;

static void prs_sma_init(void)
{
	if (prs_sma_region)
		return;

	/* waste about 400k for sma... */
	prs_sma_region = sma_alloc_region(16384, 24);
}

void prs_sma_shutdown(void)
{
	if (!prs_sma_region)
		return;

	if (sma_free_region(prs_sma_region) == -1)
	{
		DEBUG(0, ("prs_sma_shutdown failed!\n"));
	}
	else
		prs_sma_region = NULL;
}


/*******************************************************************
dump a prs to a file
 ********************************************************************/
void prs_dump(char *name, int v, prs_struct *ps)
{
	int fd, i;
	pstring fname;
	if (!DEBUGLVL(50))
		return;
	for (i=1;i<100;i++) {
		if (v != -1) {
			slprintf(fname,sizeof(fname), "/tmp/%s_%d.%d.prs", name, v, i);
		} else {
			slprintf(fname,sizeof(fname), "/tmp/%s.%d.prs", name, i);
		}
		fd = open(fname, O_WRONLY|O_CREAT|O_EXCL, 0644);
		if (fd != -1 || errno != EEXIST) break;
	}
	if (fd != -1) {
		write(fd, ps->data, ps->data_size);
		close(fd);
		DEBUG(0,("created %s\n", fname));
	}
}

/*******************************************************************
 debug output for parsing info.
 ********************************************************************/
void prs_debug(prs_struct *ps, int depth, const char *desc,
	       const char *fn_name)
{
	const char *marshall_info = "";
	const char *colon = "";

	CHECK_STRUCT(ps);
	if (depth == -1)
		depth = ps->depth;
	if (depth == 0)
	{
		if (MARSHALLING(ps))
			marshall_info = "marshalling";
		else
			marshall_info = "unmarshalling";

		if (desc && desc[0])
			colon = ": ";

		safe_free(ps->last_level0);
		ps->last_level0 = NULL;
		asprintf(&ps->last_level0, "%s %s%s%s", fn_name,
			 marshall_info, colon, desc);

		/* Only create a header at depth 0 */
		if (! DEBUGLVL(5 + depth))
			return;
	}
	DEBUGADD(5 + depth, ("%s%06x %s %s%s%s\n",
			     tab_depth(depth), ps->offset, fn_name,
			     marshall_info, colon, desc));
}

/*******************************************************************
 debug a parse structure
 ********************************************************************/
void prs_debug_out(const prs_struct *ps, char *msg, int level)
{
	CHECK_STRUCT(ps);

	if (!DEBUGLVL(level))
		return;

	DEBUG(level, ("%s ps: io %s align %d offset %d err %d data %p\n",
		      msg, BOOLSTR(ps->io), ps->align, ps->offset, ps->error,
		      ps->data));

	if (ps->data != NULL)
	{
		dump_data(level, ps->data, prs_buf_len(ps));
	}
}

/*******************************************************************
 initialise a parse structure
 ********************************************************************/
BOOL prs_init(prs_struct *ps, uint32 size, uint8 align, BOOL io)
{
	ps->struct_start = 0xfefefefe;
	ps->io = io;
	ps->align = align;
	ps->align_fn = NULL;
	ps->offset = 0;
	ps->error = False;
	ps->bigendian = False;
	ps->side = PRS_CLIENT;

	ps->data = NULL;
	ps->data_size = 0;

	ps->start = 0;
	ps->end = 0;

	ps->pending_stack = NULL;
	ps->parse_error = 0;
	ps->depth = 0;
	ps->last_level0 = NULL;

	ps->struct_end = 0xdcdcdcdc;

	if (size != 0)
	{
		prs_realloc_data(ps, size);
		ps->end = 0xffffffff;
	}

	CHECK_STRUCT(ps);
	return True;
}

static uint32 pointer_token;

/*******************************************************************
 create a unique, non-zero pointer token
 ********************************************************************/
void prs_reset_ptr_token(void)
{
	pointer_token = 0;
}

/*******************************************************************
 create a unique, non-zero pointer token
 ********************************************************************/
uint32 prs_ptr_token(void)
{
	pointer_token++;
	return pointer_token;
}

/*******************************************************************
 set the packed data representation type in a parse structure
 ********************************************************************/
void prs_set_packtype(prs_struct *ps, const uint8 *pack_type)
{
	CHECK_STRUCT(ps);
	ps->bigendian = pack_type[0] == 0x0;
	DEBUG(10, ("prs_set_packtype: bigendian: %s\n",
		   BOOLSTR(ps->bigendian)));
}

/*******************************************************************
 create a parse structure
 ********************************************************************/
void prs_create(prs_struct *ps, const char *data, uint32 size,
		uint8 align, BOOL io)
{
	DEBUG(200, ("prs_create: data:%p size:%d align:%d io:%s\n",
		    data, size, align, BOOLSTR(io)));

	prs_init(ps, 0, align, io);
	ps->data = data;
	ps->data_size = size;
	ps->end = ps->start + size;

	CHECK_STRUCT(ps);
}

/*******************************************************************
 copy a parse structure
 ********************************************************************/
BOOL prs_copy(prs_struct *ps, const prs_struct *from)
{
	int len = prs_buf_len(from);
	CHECK_STRUCT(ps);
	CHECK_STRUCT(from);
	prs_init(ps, len, from->align, from->io);
	ps->bigendian = from->bigendian;
	if (len != 0)
	{
		if (ps->data == NULL)
		{
			return False;
		}
		if (!prs_buf_copy(ps->data, from, 0, len))
		{
			return False;
		}
	}
	ps->offset = len;
	prs_set_start(ps, 0);
	return True;
}

/*******************************************************************
 allocate a memory buffer.  assume it's empty
 ********************************************************************/
static BOOL prs_alloc_data(prs_struct *buf, int size)
{
	prs_sma_init();

	CHECK_STRUCT(buf);

	buf->data_size = size;
	buf->data = sma_alloc(prs_sma_region, buf->data_size);

	if (buf->data == NULL && size != 0)
	{
		DEBUG(3, ("prs_alloc: could not malloc size %d\n",
			  buf->data_size));
		buf->data_size = 0;

		return False;
	}

	memset(buf->data, 0, buf->data_size);
	buf->end = buf->start + size;

	CHECK_STRUCT(buf);
	return True;
}

/*******************************************************************
 search for a memory buffer that falls within the specified offset
 ********************************************************************/
static const prs_struct *prs_find(const prs_struct *buf, uint32 offset)
{
	return buf;
}

/*******************************************************************
 allocates a memory buffer structure
 ********************************************************************/
BOOL prs_buf_copy(char *copy_into, const prs_struct *buf,
		  uint32 offset, uint32 len)
{
	uint32 end = offset + len;
	char *q = NULL;
	uint32 start_offset = offset;
	const prs_struct *bcp = buf;

	if (buf == NULL || copy_into == NULL)
		return False;

	CHECK_STRUCT(buf);
	if (DEBUGLVL(200))
	{
		DEBUG(200, ("prs_struct_copy: data[%d..%d] "
			"offset %d len %d\n",
		       buf->start, prs_buf_len(buf), offset, len));
	}

	prs_debug_out(bcp, "prs_struct_copy", 200);

	while (offset < end && ((q = prs_data(bcp, offset)) != NULL))
	{
		uint32 copy_len;
		bcp = prs_find(bcp, offset);
		copy_len = bcp->end - offset;

		DEBUG(200, ("\tdata[%d..%d] - offset %d len %d\n",
			    bcp->start, bcp->end, offset, copy_len));

		memcpy(copy_into, q, copy_len);

		offset += copy_len;
		copy_into += copy_len;
	}

	if (bcp != NULL)
	{
		DEBUG(200, ("prs_struct_copy: copied %d bytes\n",
			    offset - start_offset));
	}
	else
	{
		DEBUG(200, ("prs_struct_copy: failed\n"));
	}

	return buf != NULL;
}

/*******************************************************************
 frees up a memory buffer.
 ********************************************************************/
void prs_struct_free(prs_struct **buf)
{
	if (buf == NULL)
		return;
	if ((*buf) == NULL)
		return;

	CHECK_STRUCT(*buf);
	prs_free_data(*buf);	/* delete memory data */
	safe_free(*buf);		/* delete item */
	(*buf) = NULL;
}

/*******************************************************************
 frees a memory buffer chain.  assumes that all items are malloced.
 ********************************************************************/
static void prs_free_chain(prs_struct **buf)
{
	if (buf == NULL)
		return;
	if ((*buf) == NULL)
		return;

	CHECK_STRUCT(*buf);
	prs_struct_free(buf);
}

/*******************************************************************
 frees a memory buffer.
 ********************************************************************/
void prs_free_data(prs_struct *ps)
{
	if (ps == NULL)
		return;

	if (ps->data != NULL)
	{
		CHECK_STRUCT(ps);
		/* delete data in this structure */
		/* prs_sma_region should really already be initialised */
		sma_free(prs_sma_region, ps->data);
		ps->data = NULL;
	}
	ps->data_size = 0;

	if (generic_list_first(ps->pending_stack, NULL))
	{
		DEBUG(1, ("WARNING: prs_free_data: "
			  "There are still pending entries\n"));
	}
	prs_clean_pending(ps, False);

	safe_free(ps->last_level0);
	ps->last_level0 = NULL;
}

/*******************************************************************
 reallocate a memory buffer
********************************************************************/
BOOL prs_realloc_data(prs_struct *buf, size_t new_size)
{
	char *new_data;

	prs_sma_init();

	CHECK_STRUCT(buf);

	prs_debug_out(buf, "prs_realloc_data - before", 200);

	SMB_ASSERT(((ssize_t) new_size) >= 0);

	if (new_size == 0)
	{
		prs_free_data(buf);
		return True;
	}

	new_data = sma_realloc(prs_sma_region, buf->data, new_size);

	if (new_data != NULL)
	{
		if (new_size > buf->data_size)
		{
			memset(&new_data[buf->data_size], 0,
			       new_size - buf->data_size);
		}
		buf->data = new_data;
		buf->data_size = new_size;
	}
	else if (buf->data_size >= new_size)
	{
		DEBUG(3, ("prs_realloc_data: warning - "
			  "could not realloc to %d\n", new_size));
	}
	else
	{
		DEBUG(3, ("prs_realloc_data: error - "
			  "could not realloc to %d\n", new_size));

		prs_free_data(buf);
		buf->error = True;
		return False;
	}

	buf->end = buf->start + new_size;

	prs_debug_out(buf, "prs_realloc_data - after", 200);
	return True;
}

/*******************************************************************
 reallocate a memory buffer, retrospectively :-)
 ********************************************************************/
BOOL prs_grow_data(prs_struct *buf, BOOL io, size_t new_size, BOOL force_grow)
{
	if (buf == NULL)
	{
		return False;
	}

	CHECK_STRUCT(buf);

	if (new_size > buf->data_size)
	{
		if (!io || force_grow)
		{
			/* writing or force realloc */
			return prs_realloc_data(buf, new_size);
		}
		else
		{
			/*
			 * reading, there is just not that much
			 * data in the buffer
			 */
			DEBUG(1, ("prs_grow_data(%s): %d > %d\n",
				  prs_last_level0(buf),
				  new_size, buf->data_size));
			return False;
		}
	}
	return True;
}


/*******************************************************************
 add up the lengths of all sections.
 ********************************************************************/
uint32 prs_buf_len(const prs_struct *buf)
{
	int len = 0;
	CHECK_STRUCT(buf);
	len += buf->end - buf->start;
	return len;
}

/*******************************************************************
 return the memory location specified by   may return NULL.
 ********************************************************************/
char *prs_data(const prs_struct *buf, uint32 offset)
{
	CHECK_STRUCT(buf);
	buf = prs_find(buf, offset);
	if (buf != NULL)
	{
		return &(buf->data[offset - buf->start]);
	}
	return NULL;
}


/*******************************************************************
 align a pointer to a multiple of align_offset bytes.  looks like it
 will work for offsets of 0, 2 and 4...
 ********************************************************************/
BOOL prs_align(prs_struct *ps)
{
	int mod;
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	mod = ps->offset & (ps->align - 1);
	if (ps->align != 0 && mod != 0)
	{
		ps->offset += ps->align - mod;
		if (!prs_grow(ps, ps->offset))
		{
			return False;
		}
	}
	return True;
}

/*******************************************************************
 attempt, if appropriate, to grow a data buffer.

 depends on the data stream mode (io)
 ********************************************************************/
BOOL prs_grow(prs_struct *ps, uint32 new_size)
{
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	return prs_grow_data(ps, ps->io, new_size, False);
}

/*******************************************************************
 * appends data in one prs_struct to another.
 ********************************************************************/
BOOL prs_append_prs(prs_struct *ps, const prs_struct *from)
{
	return prs_append_data(ps, prs_data(from, 0), from->offset);
}

/*******************************************************************
 lengthens a buffer by len bytes and copies data into it.
 ********************************************************************/
BOOL prs_append_data(prs_struct *ps, const char *data, int len)
{
	int prev_size = ps->data_size;
	int new_size = prev_size + len;
	char *to;

	DEBUG(200, ("prs_append_data: prev_size: %d new_size: %d\n",
		    prev_size, new_size));

	CHECK_STRUCT(ps);
	prs_realloc_data(ps, new_size);
	to = prs_data(ps, prev_size);

	if (to == NULL || ps->data_size != new_size)
	{
		return False;
	}
	memcpy(to, data, len);

	return True;
}

BOOL prs_add_data(prs_struct *ps, const char *data, int len)
{
	int prev_size;
	int new_size;
	char *to = NULL;

	ps->offset = 0;

	if (ps->data == NULL)
	{
		DEBUG(10, ("prs_add_data: new_size: %d\n", len));
		prs_init(ps, len, 4, True);
		prev_size = 0;
		new_size = len;
		if (ps->data == NULL)
		{
			return False;
		}
	}
	else
	{
		prev_size = ps->data_size;
		new_size = prev_size + len;
		DEBUG(10, ("prs_add_data: prev_size: %d new_size: %d\n",
			   prev_size, new_size));
		if (!prs_realloc_data(ps, new_size))
		{
			return False;
		}
	}

	DEBUG(10, ("ps->start: %d\n", ps->start));
	ps->start = 0x0;

	to = prs_data(ps, prev_size);
	if (to == NULL)
	{
		DEBUG(10, ("prs_add_data: data could not be found\n"));
		return False;
	}
	if (ps->data_size != new_size)
	{
		DEBUG(10, ("prs_add_data: ERROR: data used %d new_size %d\n",
			   ps->data_size, new_size));
		return False;
	}
	memcpy_zero(to, data, len);
	return True;
}

/*******************************************************************
 Change the struct type.
 ********************************************************************/

void prs_switch_type(prs_struct *ps, BOOL io)
{
	if ((ps->io ^ io) == True)
		ps->io = io;
}

/*******************************************************************
 Force a prs_struct to be dynamic even when its size is 0.
 ********************************************************************/

void prs_force_dynamic(prs_struct *ps)
{
}

/*******************************************************************
 Set the start of the buffer
 ********************************************************************/
BOOL prs_set_start(prs_struct *ps, uint32 start)
{
	ps->start = start;
	ps->end = ps->start + ps->offset;
	return True;
}

/*******************************************************************
 Fetch the current offset (external interface).
 ********************************************************************/
uint32 prs_data_size(prs_struct *ps)
{
	return ps->data_size;
}

/*******************************************************************
 Fetch the current offset (external interface).
 ********************************************************************/
uint32 prs_offset(prs_struct *ps)
{
	return ps->offset;
}

/*******************************************************************
 Set the current offset (external interface).
 ********************************************************************/
BOOL prs_set_offset(prs_struct *ps, uint32 offset)
{
	if (offset <= ps->offset)
	{
		ps->offset = offset;
		return True;
	}

	if (!prs_grow(ps, offset))
	{
		return False;
	}

	ps->offset = offset;
	return True;
}

/*******************************************************************
 Set the current depth (external interface).
 ********************************************************************/
void prs_set_depth(prs_struct *ps, int depth)
{
	if (! ps)
		return;

	if (depth != -1)
		ps->depth = depth;
}

/*******************************************************************
 Modify the current depth (external interface).
 ********************************************************************/
void prs_inc_depth(prs_struct *ps)
{
	if (! ps)
		return;

	ps->depth++;
}

void prs_dec_depth(prs_struct *ps)
{
	if (! ps)
		return;

	ps->depth--;
}

/*******************************************************************
 Get the current depth (external interface).
 ********************************************************************/
int prs_depth(prs_struct *ps)
{
	if (! ps)
		return 0;

	return ps->depth;
}

/*******************************************************************
 Set the "side" this ps works for
 ********************************************************************/
void prs_set_side(prs_struct *ps, BOOL side)
{
	if (! ps)
		return;
	CHECK_STRUCT(ps);

	ps->side = side;

	DEBUG(10, ("prs_set_side(%s)\n",
		   (PRS_IS_CLIENT(ps) ? "client" : "server")));
}

/*******************************************************************
 set/get the last parse error
 Yes, this is DCE/RPC specific
 ********************************************************************/
void prs_set_parse_error(prs_struct *ps, uint32 parse_error)
{
	if (! ps)
		return;
	CHECK_STRUCT(ps);

	ps->parse_error = parse_error;

	DEBUG(5, ("prs_set_parse_error(0x%x)\n", parse_error));
}

uint32 prs_parse_error(const prs_struct *ps)
{
	if (! ps)
		return 0x0;

	return ps->parse_error;
}


const char *prs_last_level0(const prs_struct *ps)
{
	if (!ps || !ps->last_level0)
		return "<NULL>";
	return ps->last_level0;
}


/*******************************************************************
 ********************************************************************/
typedef struct _PendingItem
{
	GenericParseCB fn;
	char *name;
	void *item;
	int depth;
} PendingItem;

typedef struct _PendingStackItem
{
	char *name;
	GENERIC_LIST *fns;
} PendingStackItem;


BOOL prs_start_pending(prs_struct *ps, const char *comment)
{
	PendingStackItem *st;

	CHECK_STRUCT(ps);

	prs_debug(ps, -1, comment, "prs_start_pending:");
	prs_inc_depth(ps);

	if (! ps->pending_stack)
		ps->pending_stack = generic_list_new();

	if (! ps->pending_stack)
	{
		DEBUG(0, ("WARNING: prs_start_pending failed\n"));
		return False;
	}

	st = g_new(PendingStackItem, 1);

	if (! st)
	{
		DEBUG(0, ("WARNING: prs_start_pending: No mem for st\n"));
		return False;
	}

	st->fns = generic_list_new();
	if (!st->fns)
		return False;

	st->name = comment ? strdup(comment) : NULL;

	return generic_list_prepend(ps->pending_stack, st, 0);
}

BOOL prs_add_pending(prs_struct *ps, GenericParseCB fn,
		     const char *name, void *item)
{
	PendingStackItem *psi;
	PendingItem *pi;

	CHECK_STRUCT(ps);

	psi = generic_list_first(ps->pending_stack, NULL);
	if (! psi)
	{
		DEBUG(0, ("WARNING: prs_add_pending(%s) called without "
			  "prs_start_pending\n", name));
		return False;
	}

	pi = g_new(PendingItem, 1);
	if (! pi)
	{
		DEBUG(0, ("WARNING: prs_add_pending(%s): No memory\n", name));
		return False;
	}

	pi->fn = fn;
	pi->name = name ? strdup(name) : NULL;
	pi->item = item;
	pi->depth = ps->depth;

	prs_debug(ps, -1, name, "add_pending:");

	return generic_list_append(psi->fns, pi, 0);
}

static void prs_clean_pending(prs_struct *ps, BOOL show)
{
	PendingStackItem *psi;
	PendingItem *pi;
	GENERIC_LIST *l;

	while((psi = generic_list_first(ps->pending_stack, NULL)))
	{
		generic_list_remove(ps->pending_stack, psi, NULL);

		l = psi->fns;

		while ((pi = generic_list_first(l, NULL)))
		{
			generic_list_remove(l, pi, NULL);

			if (show)
			{
				ps->depth = pi->depth;

				prs_debug(ps, -1, pi->name, "missing:");
			}

			safe_free(pi->name);
			safe_free(pi);
		}

		generic_list_destroy(l);

		safe_free(psi->name);
		safe_free(psi);
	}

	generic_list_destroy(ps->pending_stack);
	ps->pending_stack = NULL;
}

BOOL prs_stop_pending(prs_struct *ps)
{
	PendingStackItem *psi;
	PendingItem *pi;
	GENERIC_LIST *l;
	int old_depth;
	BOOL ret = True;

	CHECK_STRUCT(ps);

	psi = generic_list_first(ps->pending_stack, NULL);
	if (! psi)
	{
		DEBUG(0, ("WARNING: prs_stop_pending without start\n"));
		return False;
	}

	prs_dec_depth(ps);
	prs_debug(ps, -1, psi->name, "prs_stop_pending starting:");

	old_depth = ps->depth;

	l = psi->fns;

	while ((pi = generic_list_first(l, NULL)))
	{
		generic_list_remove(l, pi, NULL);

		ps->depth = pi->depth;

		prs_debug(ps, -1, pi->name, "pending_do:");

		if (pi->fn)
		{
			ret = pi->fn(pi->name, pi->item, ps);
		}
		else
		{
			prs_debug(ps, -1, pi->name, "NULL");
			ret = True;
		}

		if (! ret)
		{
			DEBUG(5, ("WARNING: parse for %s failed\n", pi->name));
		}

		safe_free(pi->name);
		free(pi);

		if (! ret)
		{
			break;
		}
	}

	if (! ret)
	{
		/* show the missing stuff */
		prs_clean_pending(ps, True);
	}

	ps->depth = old_depth;

	if (ret)
	{
		/* just clean the list up */
		generic_list_remove(ps->pending_stack, psi, NULL);
		generic_list_destroy(psi->fns);
		psi->fns = NULL;
		prs_debug(ps, -1, psi->name, "prs_stop_pending finished:");
		safe_free(psi->name);
		free(psi);
	}

	return ret;
}

/*******************************************************************
 Delete the memory in a parse structure - if we own it.
 ********************************************************************/

void prs_mem_free(prs_struct *ps)
{
	prs_free_data(ps);
	ps->offset = 0;
}


/*******************************************************************
 Append some data from one parse_struct into another.
 ********************************************************************/

BOOL prs_append_some_prs_data(prs_struct *dst, prs_struct *src, int32 start,
			      uint32 len)
{
	char *src_data;

	/* 
	 * JFM:
	 * ok, it looks like shit. It smells like shit.
	 * summary: it's shit.
	 *
	 * I'm not proud of that code. I mae that ugly hack to
	 * a) not change luke's prs memory managment
	 * b) not change the spoolss parsing code
	 *
	 * mail me, call me, hit me before changing that piece of code.
	 */

	src_data = prs_data(src, start);
	if (start == 0)
	{
		prs_add_data(dst, src_data, len);
	}
	else
	{
		dst->data_size = 0;
		prs_add_data(dst, src_data, len);
	}

	dst->offset = dst->data_size;
	return True;
}

/*******************************************************************
 stream a uint8
 ********************************************************************/
BOOL _prs_uint8(char *name, prs_struct *ps, int depth, uint8 *data8)
{
	char *q;
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	if (depth == -1)
		depth = ps->depth;
	if (!prs_grow(ps, ps->offset + 1))
	{
		return False;
	}
	q = prs_data(ps, ps->offset);
	if (q == NULL)
	{
		ps->error = True;
		prs_debug_out(ps, "_prs_uint8 error", 5);
		return False;
	}

	DBG_RW_CVAL(name, depth, ps->offset, ps->io, q, *data8);
	ps->offset += 1;

	return True;
}

/*******************************************************************
 stream a uint16
 ********************************************************************/
BOOL _prs_uint16(char *name, prs_struct *ps, int depth, uint16 *data16)
{
	char *q;
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	PRS_ALIGN(ps, 2);
	if (depth == -1)
		depth = ps->depth;
	if (!prs_grow(ps, ps->offset + 2))
	{
		return False;
	}
	q = prs_data(ps, ps->offset);
	if (q == NULL)
	{
		ps->error = True;
		prs_debug_out(ps, "_prs_uint16 error", 5);
		return False;
	}

	DBG_RW_SVAL(name, depth, ps->offset, ps->io, ps->bigendian, q,
		    *data16);
	ps->offset += 2;

	return True;
}

/*******************************************************************
 hash a stream.
 ********************************************************************/
BOOL _prs_hash1(prs_struct *ps, uint32 offset, uint8 sess_key[16])
{
	char *q;
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	q = prs_data(ps, ps->offset);
	if (q == NULL)
	{
		ps->error = True;
		prs_debug_out(ps, "_prs_hash1 error", 5);
		return False;
	}

#ifdef DEBUG_PASSWORD
	DEBUG(100, ("prs_hash1\n"));
	dump_data(100, sess_key, 16);
	dump_data(100, q, 68);
#endif
	SamOEMhash((uchar *) q, sess_key, 2);
#ifdef DEBUG_PASSWORD
	dump_data(100, q, 68);
#endif

	return True;
}

/*******************************************************************
 stream a uint32
 ********************************************************************/
BOOL _prs_uint32(char *name, prs_struct *ps, int depth, uint32 *data32)
{
	char *q;
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	PRS_ALIGN(ps, 4);
	if (depth == -1)
		depth = ps->depth;
	if (!prs_grow(ps, ps->offset + 4))
	{
		return False;
	}
	q = prs_data(ps, ps->offset);
	if (q == NULL)
	{
		fstring str;
		slprintf(str, sizeof(str) - 1, "_prs_uint32 error (%s)",
			 name);
		ps->error = True;
		prs_debug_out(ps, str, 5);
		return False;
	}

	DBG_RW_IVAL(name, depth, ps->offset, ps->io, ps->bigendian, q,
		    *data32);
	ps->offset += 4;

	return True;
}


/******************************************************************
 stream an array of uint8s.  length is number of uint8s
 ********************************************************************/
BOOL _prs_uint8s(BOOL charmode, char *name, prs_struct *ps, int depth,
		 uint8 *data8s, int len)
{
	char *q;
	int end_offset;
	char *e;
	CHECK_STRUCT(ps);

	if (len == 0)
	{
		return True;
	}

	if (ps->error)
		return False;
	if (depth == -1)
		depth = ps->depth;
	end_offset = ps->offset + len * sizeof(uint8);
	if (!prs_grow(ps, end_offset))
	{
		return False;
	}
	q = prs_data(ps, ps->offset);
	e = prs_data(ps, end_offset - 1);

	if (q == NULL || e == NULL)
	{
		ps->error = True;
		prs_debug_out(ps, "_prs_uint8s error", 5);
		return False;
	}

	DBG_RW_PCVAL(charmode, name, depth, ps->offset, ps->io, q, data8s,
		     len);
	ps->offset = end_offset;

	return True;
}

/******************************************************************
 stream an array of uint16s.  length is number of uint16s
 ********************************************************************/
BOOL _prs_uint16s(BOOL charmode, char *name, prs_struct *ps, int depth,
		  uint16 *data16s, int len)
{
	char *q;
	int end_offset;
	char *e;
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	PRS_ALIGN(ps, 2);
	if (depth == -1)
		depth = ps->depth;

	if (len == 0)
	{
		return True;
	}

	end_offset = ps->offset + len * sizeof(uint16);
	if (!prs_grow(ps, end_offset))
	{
		return False;
	}
	q = prs_data(ps, ps->offset);
	e = prs_data(ps, end_offset - 1);

	if (q == NULL || e == NULL)
	{
		ps->error = True;
		prs_debug_out(ps, "_prs_uint16s error", 5);
		return False;
	}

	DBG_RW_PSVAL(charmode, name, depth, ps->offset, ps->io, ps->bigendian,
		     q, data16s, len);
	ps->offset = end_offset;

	return True;
}

/******************************************************************
 stream an array of uint32s.  length is number of uint32s
 ********************************************************************/
BOOL _prs_uint32s(BOOL charmode, char *name, prs_struct *ps, int depth,
		  uint32 *data32s, int len)
{
	char *q;
	int end_offset;
	char *e;
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	PRS_ALIGN(ps, 4);
	if (depth == -1)
		depth = ps->depth;

	if (len == 0)
	{
		return True;
	}

	end_offset = ps->offset + len * sizeof(uint32);
	if (!prs_grow(ps, end_offset))
	{
		return False;
	}
	q = prs_data(ps, ps->offset);
	e = prs_data(ps, end_offset - 1);

	if (q == NULL || e == NULL)
	{
		ps->error = True;
		prs_debug_out(ps, "_prs_uint32s error", 5);
		return False;
	}

	DBG_RW_PIVAL(charmode, name, depth, ps->offset, ps->io, ps->bigendian,
		     q, data32s, len);
	ps->offset = end_offset;

	return True;
}

/******************************************************************
 stream a "not" unicode string, length/buffer specified separately,
 in byte chars
 ********************************************************************/
BOOL _prs_buffer2(BOOL charmode, char *name, prs_struct *ps, int depth,
		  BUFFER2 * str)
{
	char *q;
	int end_offset;
	char *e;
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	if (depth == -1)
		depth = ps->depth;

	if (str->buf_len == 0)
	{
		return True;
	}

	end_offset = ps->offset + str->buf_len * sizeof(uint8);
	if (!prs_grow(ps, end_offset))
	{
		return False;
	}
	q = prs_data(ps, ps->offset);
	e = prs_data(ps, end_offset - 1);

	if (q == NULL || e == NULL)
	{
		ps->error = True;
		prs_debug_out(ps, "_prs_buffer2 error", 5);
		return False;
	}

	DBG_RW_PCVAL(charmode, name, depth, ps->offset, ps->io, q,
		     str->buffer, str->buf_len);
	ps->offset = end_offset;

	return True;
}

/******************************************************************
 stream a string, length/buffer specified separately,
 in uint8 chars.
 ********************************************************************/
BOOL _prs_string2(BOOL charmode, char *name, prs_struct *ps, int depth,
		  STRING2 * str)
{
	char *q;
	int end_offset;
	char *e;
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	if (depth == -1)
		depth = ps->depth;

	if (str->str_str_len == 0)
	{
		return True;
	}

	end_offset = ps->offset + str->str_str_len * sizeof(uint8);
	if (!prs_grow(ps, end_offset))
	{
		return False;
	}
	q = prs_data(ps, ps->offset);
	e = prs_data(ps, end_offset - 1);

	if (q == NULL || e == NULL)
	{
		ps->error = True;
		prs_debug_out(ps, "_prs_string2 error", 5);
		return False;
	}

	DBG_RW_PCVAL(charmode, name, depth, ps->offset, ps->io, q,
		     str->buffer, str->str_str_len);
	ps->offset = end_offset;

	return True;
}

/******************************************************************
 stream a unicode string, length/buffer specified separately,
 in uint16 chars.
 ********************************************************************/
BOOL _prs_unistr2(BOOL charmode, char *name, prs_struct *ps, int depth,
		  UNISTR2 *str)
{
	char *q;
	int end_offset;
	char *e;
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	if (depth == -1)
		depth = ps->depth;

	if (str->uni_str_len == 0)
	{
		return True;
	}

	end_offset = ps->offset + str->uni_str_len * sizeof(uint16);
	if (!prs_grow(ps, end_offset))
	{
		return False;
	}
	q = prs_data(ps, ps->offset);
	e = prs_data(ps, end_offset - 1);

	if (q == NULL || e == NULL)
	{
		ps->error = True;
		prs_debug_out(ps, "_prs_unistr2 error", 5);
		return False;
	}

	DBG_RW_PSVAL(charmode, name, depth, ps->offset, ps->io, ps->bigendian,
		     q, str->buffer, str->uni_str_len);
	ps->offset = end_offset;

	return True;
}

/******************************************************************
 stream a unicode string, length/buffer specified separately,
 in uint16 chars.
 ********************************************************************/
BOOL prs_unistr3(BOOL charmode, char *name, UNISTR3 * str,
		 prs_struct *ps, int depth)
{
	char *q;
	int end_offset;
	char *e;
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	if (depth == -1)
		depth = ps->depth;

	if (str->uni_str_len == 0)
	{
		return True;
	}

	end_offset = ps->offset + str->uni_str_len * sizeof(uint16);
	if (!prs_grow(ps, end_offset))
	{
		return False;
	}
	q = prs_data(ps, ps->offset);
	e = prs_data(ps, end_offset - 1);

	if (q == NULL || e == NULL)
	{
		ps->error = True;
		prs_debug_out(ps, "_prs_unistr3 error", 5);
		return False;
	}

	DBG_RW_PSVAL(charmode, name, depth, ps->offset, ps->io, ps->bigendian,
		     q, str->str.buffer, str->uni_str_len);
	ps->offset = end_offset;

	return True;
}

/*******************************************************************
 stream a unicode  null-terminated string
 ********************************************************************/
BOOL _prs_unistr(char *name, prs_struct *ps, int depth, UNISTR * str)
{
	int i = -1;
	uint8 *start;
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	if (depth == -1)
		depth = ps->depth;
	start = (uint8 *)prs_data(ps, ps->offset);

	do
	{
		char *q;
		i++;
		if (!prs_grow(ps, ps->offset + (i + 1) * 2))
		{
			return False;
		}
		q = prs_data(ps, ps->offset + i * 2);
		if (q == NULL)
		{
			ps->error = True;
			prs_debug_out(ps, "_prs_unistr error", 5);
			return False;
		}
		RW_SVAL(ps->io, ps->bigendian, q, str->buffer[i], 0);
	}
	while ((((size_t) i) < sizeof(str->buffer) / sizeof(str->buffer[0]))
	       && (str->buffer[i] != 0));


	ps->offset += (i + 1) * 2;

	dump_data(5 + depth, (char *)start, i * 2);

	return True;
}

/*******************************************************************
 stream a null-terminated string.  len is strlen, and therefore does
 not include the null-termination character.

 len == 0 indicates variable length string
 (up to max size of pstring - 1024 chars).

 ********************************************************************/
BOOL _prs_string(char *name, prs_struct *ps, int depth, char *str,
		 uint16 len, uint16 max_buf_size)
{
	int i = -1;		/* start off at zero after 1st i++ */
	BOOL len_limited;

	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	if (depth == -1)
		depth = ps->depth;

	len_limited = len == 0 || !ps->io;

	DEBUG(200, ("_prs_string: string %s len %d max %d\n",
		    str, len, max_buf_size));

	DEBUG(5 + depth,
	      ("%s%04x %s: ", tab_depth(depth), ps->offset,
	       name != NULL ? name : ""));
	do
	{
		char *q;
		i++;

		if (!prs_grow(ps, ps->offset + i + 1))
		{
			return False;
		}
		q = prs_data(ps, ps->offset + i);
		if (q == NULL)
		{
			ps->error = True;
			DEBUG(5 + depth, ("%s\n", str));
			prs_debug_out(ps, "_prs_string error", 5);
			return False;
		}

		if (i < len || len_limited)
		{
			RW_CVAL(ps->io, q, str[i], 0);
		}
		else
		{
			uint8 dummy = 0;
			RW_CVAL(ps->io, q, dummy, 0);
		}

	}
	while (i < max_buf_size && (len_limited ? str[i] != 0 : i < len));

	ps->offset += i + 1;

	DEBUG(5 + depth, ("%s\n", str));

	return True;
}

/*******************************************************************
 prs_uint16 wrapper.  call this and it sets up a pointer to where the
 uint16 should be stored, or gets the size if reading
 ********************************************************************/
BOOL _prs_uint16_pre(char *name, prs_struct *ps, int depth, uint16 *data16,
		     uint32 *offset)
{
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	if (depth == -1)
		depth = ps->depth;
	(*offset) = ps->offset;
	if (ps->io)
	{
		/* reading. */
		return _prs_uint16(name, ps, depth, data16);
	}
	else
	{
		PRS_ALIGN(ps, 2);
		ps->offset += sizeof(uint16);
	}
	return True;
}

/*******************************************************************
 prs_uint16 wrapper.  call this and it retrospectively stores the size.
 does nothing on reading, as that is already handled by ...._pre()
 ********************************************************************/
BOOL _prs_uint16_post(char *name, prs_struct *ps, int depth, uint16 *data16,
		      uint32 ptr_uint16, uint32 start_offset)
{
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	if (depth == -1)
		depth = ps->depth;
	if (!ps->io)
	{
		/* storing: go back and do a retrospective job.  i hate this */
		uint16 data_size = ps->offset - start_offset;
		uint32 old_offset = ps->offset;

		ps->offset = ptr_uint16;
		prs_uint16(name, ps, depth, &data_size);
		ps->offset = old_offset;
	}
	else
	{
		ps->offset = start_offset + (*data16);
	}
	return True;
}

/*******************************************************************
 prs_uint32 wrapper.  call this and it sets up a pointer to where the
 uint32 should be stored, or gets the size if reading
 ********************************************************************/
BOOL _prs_uint32_pre(char *name, prs_struct *ps, int depth, uint32 *data32,
		     uint32 *offset)
{
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	if (depth == -1)
		depth = ps->depth;
	(*offset) = ps->offset;
	if (ps->io)
	{
		/* reading. */
		return _prs_uint32(name, ps, depth, data32);
	}
	else
	{
		PRS_ALIGN(ps, 4);
		ps->offset += sizeof(uint32);
	}
	return True;
}

/*******************************************************************
 prs_uint32 wrapper.  call this and it retrospectively stores the size.
 does nothing on reading, as that is already handled by ...._pre()
 ********************************************************************/
BOOL _prs_uint32_post(char *name, prs_struct *ps, int depth, uint32 *data32,
		      uint32 ptr_uint32, uint32 data_size)
{
	CHECK_STRUCT(ps);
	if (ps->error)
		return False;
	if (depth == -1)
		depth = ps->depth;
	if (!ps->io)
	{
		/* storing: go back and do a retrospective job.  i hate this */
		uint32 old_offset = ps->offset;
		ps->offset = ptr_uint32;
		prs_uint32(name, ps, depth, &data_size);
		ps->offset = old_offset;
	}
	else
	{
		ps->offset = data_size + (*data32);
	}
	return True;
}

/*******************************************************************
 prs_tdb_store.  stores prs_struct data by prs_struct key 
 ********************************************************************/
int prs_tdb_delete(TDB_CONTEXT * tdb, prs_struct *pk)
{
	TDB_DATA key;

	key.dptr = (char *)prs_data(pk, 0);
	key.dsize = prs_buf_len(pk);

	return tdb_delete(tdb, key);
}

/*******************************************************************
 prs_tdb_store.  stores prs_struct data by prs_struct key 
 ********************************************************************/
int prs_tdb_store(TDB_CONTEXT * tdb, int flgs, prs_struct *pk, prs_struct *pd)
{
	TDB_DATA key;
	TDB_DATA data;

	key.dptr = (char *)prs_data(pk, 0);
	key.dsize = prs_buf_len(pk);

	data.dptr = prs_data(pd, 0);
	data.dsize = prs_buf_len(pd);

	return tdb_store(tdb, key, data, flgs);
}

/*******************************************************************
 prs_tdb_fetch.  fetches prs_struct data by prs_struct key 
 ********************************************************************/
void prs_tdb_fetch(TDB_CONTEXT * tdb, prs_struct *pk, prs_struct *pd)
{
	TDB_DATA key;
	TDB_DATA data;

	key.dptr = (char *)prs_data(pk, 0);
	key.dsize = prs_buf_len(pk);

	data = tdb_fetch(tdb, key);

	prs_init(pd, 0, 4, True);
	prs_append_data(pd, data.dptr, data.dsize);
	safe_free(data.dptr);
}


/* useful function to store a structure in rpc wire format */
int tdb_prs_store(TDB_CONTEXT *tdb, char *keystr, prs_struct *ps)
{
	TDB_DATA kbuf, dbuf;
	kbuf.dptr = keystr;
	kbuf.dsize = strlen(keystr)+1;
	dbuf.dptr = (char *)prs_data(ps, 0);
	dbuf.dsize = prs_buf_len(ps);

	return tdb_store(tdb, kbuf, dbuf, TDB_REPLACE);
}

/* useful function to fetch a structure into rpc wire format */
int tdb_prs_fetch(TDB_CONTEXT *tdb, char *keystr, prs_struct *ps)
{
	TDB_DATA kbuf, dbuf;
	kbuf.dptr = keystr;
	kbuf.dsize = strlen(keystr)+1;

	dbuf = tdb_fetch(tdb, kbuf);
	if (!dbuf.dptr) return -1;

	ZERO_STRUCTP(ps);
	prs_init(ps, 0, 4, UNMARSHALL);
	prs_append_data(ps, dbuf.dptr, dbuf.dsize);
	safe_free(dbuf.dptr);

	return 0;
}

