/*
 * mkproto.c - dig out function prototypes from C files
 * Copyright (C) 2001 Peter Samuelson
 *
 * The design is based on the awk program 'mkproto.awk' and its shell
 * wrapper 'mkproto.sh', both by the Samba Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
 * USA.
 */

/* the ifdefs are based on what is in include/includes.h */
#include "config.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>

#ifdef HAVE_SYS_UNISTD_H
#include <sys/unistd.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#ifndef HAVE_STRCHR
#define strchr index
#define strrchr rindex
#endif

#define BUFLEN 4096
#define LINELEN 512	/* would you believe some lines are >256 char? */
#define HEADER_TEMP ".header.tmp"
#define PROTO_TEMP ".proto.tmp"

#define STRUCT_SIZE(s) (sizeof(s) / sizeof(*(s)))
#define ZERO_MTIME(st) (st.st_mtime = 0)
#define NEWER(s1,s2) (s1.st_mtime > s2.st_mtime)

static char *progname;
static char *header_def = "_PROTO_H_";
static char *header_file;
static char *source_dir = ".";
static char *proto_dir = ".proto";
static int force;

static char *start_tokens[] = {
  /* common culprits, in order of popularity */
  "BOOL", "void", "uint32", "int", "char", "struct",
  "enum winbindd_result", "smb_ucs2_t", "ssize_t", "const",
  "size_t", "files_struct", "time_t", "LOCAL_GRP",

  /* uncommon culprits, shortest-to-longest */
  "DIR", "FILE", "long", "pid_t", "off_t", "int32", "ino_t",
  "mode_t", "uint16", "DOM_SID", "UNISTR2", "unsigned", "TDB_DATA",
  "SMB_OFF_T", "DOMAIN_GRP", "JOB_INFO_2", "JOB_INFO_1",
  "TALLOC_CTX", "TDB_CONTEXT", "user_struct", "SMB_BIG_UINT",
  "pipes_struct", "hash_element", "GENERIC_LIST", "NT_DEVICEMODE",
  "PRINTER_INFO_2", "PRINTER_INFO_1", "cli_connect_fns",
  "connection_struct", "msrpc_service_fns", "SMB_STRUCT_DIRENT",
  "enum protocol_types", "enum remote_arch_types",
  "NT_USER_TOKEN", "smb_iconv_t", "SEC_ACL",
  "NTSTATUS", "WERROR",
};

static char *ignore_start_tokens[] = {
  "static", "extern",
};

static struct { char *token, *fmt; } loadparm_tokens[] = {
  /* once again in order of popularity */
  { "FN_GLOBAL_STRING",  "char *%s(void)" },
  { "FN_GLOBAL_BOOL",    "BOOL %s(void)" },
  { "FN_LOCAL_BOOL",     "BOOL %s(int )" },
  { "FN_LOCAL_STRING",   "char *%s(int )" },
  { "FN_GLOBAL_INTEGER", "int %s(void)" },
  { "FN_LOCAL_INTEGER",  "int %s(int )" },
  { "FN_VUSER_STRING",   "char *%s(const user_struct* )" },
  { "FN_LOCAL_CHAR",     "char %s(int )" },
};

static void usage (void)
{
  fprintf (stderr,
	   "usage: %s [-h header_def] output_header input_file...\n",
	   progname);
  exit (1);
}

static char *chompsp (char *str)
{
  char *n = strchr (str, '\n');
  while (n >= str && isspace (*n)) {
    *n-- = '\0';
  }
  return str;
}

static void report_inprogress (char *filename)
{
  fprintf (stderr, "creating %s... ", filename);
}

static void report_finished (int changed)
{
  fprintf (stderr, "%s.\n", changed ? "done" : "unchanged");
}

static FILE *xfopen (char *path, char *mode)
{
  FILE *ret = fopen (path, mode);
  if (!ret) {
    char buf[LINELEN];
    sprintf (buf, "%s: cannot open %s", progname, path);
    perror (buf);
    exit (1);
  }
  return ret;
}

/*
 * mkdir_p: mkdir() but also create needed parent dirs, like mkdir(1) -p
 * This writes to 'dir' string, so DO NOT pass in a string constant.
 * For arguments, return value and errno values, see mkdir(2)
 */
static int mkdir_p (char *dir, mode_t mode)
{
  int level = 0;
  char *d;

  while (1) {
    struct stat st;
    if (!stat (dir, &st) && (st.st_mode & S_IFDIR)) {
      for (d=dir; level; level--) {
	d += strlen(d);
	*d = '/';
	if (mkdir (dir, mode))
	  return -1;
      }
      return 0;
    }
    d = strrchr (dir, '/');
    if (d) {
      *d = '\0';
      level++;
    } else {
      if (mkdir (dir, mode))
	return -1;
    }
  }
}


#ifndef HAVE_RENAME
/* from lib/replace.c, originally from GNU libiberty */
static int rename (const char *zfrom, const char *zto)
{
  if (link (zfrom, zto) < 0) {
    if (errno != EEXIST)
      return -1;
    if (unlink (zto) < 0 ||
	link (zfrom, zto) < 0)
      return -1;
  }
  return unlink (zfrom);
}
#endif /* HAVE_RENAME */


static int move_if_change (char *temp, char *file)
{
  struct stat s_t, s_f;
  static char buf[LINELEN];

  ZERO_MTIME (s_t); ZERO_MTIME (s_f);
  stat (temp, &s_t);
  stat (file, &s_f);
  if (NEWER (s_t, s_f)) {
    FILE *f_t, *f_f;
    if (s_t.st_size != s_f.st_size)
      goto do_move;
    f_t = xfopen (temp, "r");
    f_f = fopen (file, "r");
    if (!f_f) {
      fclose (f_t);
      goto do_move;
    }
    while (!feof (f_t)) {
      char buf1[BUFLEN], buf2[BUFLEN];
      int n1 = fread (buf1, 1, BUFLEN, f_t);
      int n2 = fread (buf2, 1, n1, f_f);
      if (n1 != n2 || n1 < 0 || memcmp (buf1, buf2, n1)) {
	fclose (f_t);
	fclose (f_f);
	goto do_move;
      }
    }
    fclose (f_t);
    fclose (f_f);
  }
  return 0;	/* didn't move file */

 do_move:
  if (s_f.st_mtime)
    unlink (file);
  else {
    char *p = strrchr (file, '/');
    if (p) {
      *p = '\0';
      if (mkdir_p (file, 0777)) {
	sprintf (buf, "%s: cannot make directory %s", progname, file);
	perror (buf);
	exit (1);
      }
      *p = '/';
    }
  }
  if (rename (temp, file)) {
    sprintf (buf, "%s: cannot rename %s to %s", progname, temp, file);
    perror (buf);
    exit (1);
  }
  return 1;	/* moved file */
}




static FILE *open_header (void)
{
  FILE *out = xfopen (HEADER_TEMP, "w");
  fprintf (out,
	   "#ifndef %s\n"
	   "#define %s\n"
	   "/* This file is automatically generated with \"make proto\". DO NOT EDIT */\n"
	   "\n", header_def, header_def);
  return out;
}

static void close_header (FILE *out)
{
  fprintf (out,
	   "#endif /* %s */\n", header_def);
  fclose (out);
  report_finished (move_if_change (HEADER_TEMP, header_file));
}

static int output_one_file (char *infile, FILE *out)
{
  int ret = -1;
  FILE *in;
  char buf[BUFLEN];

  fprintf (out, "\n/*The following definitions come from  %s  */\n\n",
	   infile);

  sprintf (buf, "%s/%s", proto_dir, infile);
  in = xfopen (buf, "r");
  while (!feof (in)) {
    int n1, n2;
    n1 = fread (buf, 1, BUFLEN, in);
    if (n1>0) {
      n2 = fwrite (buf, 1, n1, out);
      if (n1 != n2)
	goto do_ret;
    }
  }
  ret = 0;
 do_ret:
  fclose (in);
  return ret;
}

static int process_one_file (FILE *in, FILE *out)
{
  char line[LINELEN];
  int ignore_next = 0, in_header = 0, got_start = 0;

  while (!feof (in) && fgets (line, LINELEN, in)) {
    int i;
    char *s;

    chompsp (line);
    if (in_header) {
      int l = strlen (line);
      if (line[l-1] == ')') {
	in_header = 0;
	fprintf (out, "%s;\n", line);
	continue;
      }
      fprintf (out, "%s\n", line);
      continue;
    }

    if (ignore_next) {
      ignore_next = 0;
      continue;
    }
    if (!isalpha (line[0]) || strchr (line, ';'))
      continue;

    for (i=0; i < STRUCT_SIZE (ignore_start_tokens); i++) {
      int l;
      s = ignore_start_tokens[i];
      l = strlen (s);
      if (!strncmp (line, s, l)) {
	if (! line[l]) {
	  ignore_next = 1;
	  break;
	} else if (line[l] == '*' || isspace (line[l]))
	  break;
      }
    }
    if (i < STRUCT_SIZE (ignore_start_tokens))	/* we hit 'break' above */
      continue;

    got_start = 0;
    for (i=0; i < STRUCT_SIZE (start_tokens); i++) {
      int l;
      s = start_tokens[i];
      l = strlen (s);
      if (!strncmp (line, s, l) &&
	  (isspace (line[l]) || line[l]=='*')) {
	got_start = 1;
	break;
      }
    }
    if (! got_start) {
      /* loadparm.c wizardry */
      for (i=0; i < STRUCT_SIZE (loadparm_tokens); i++) {
	int l;
	char *e;
	s = loadparm_tokens[i].token;
	l = strlen (s);
	if (!strncmp (line, s, l) && line[l] == '(' &&
	    (e = strchr (line+l, ','))) {
	  *e = '\0';
	  fprintf (out, loadparm_tokens[i].fmt, line+l+1);
	  fprintf (out, ";\n");
	  break;
	}
      }
      continue;
    }

    s = strchr (line, '(');
    if (s) {
      int l = strlen (s);
      if (s[l-1] == ')') {
	fprintf (out, "%s;\n", line);
	continue;
      }
      in_header = 1;
      fprintf (out, "%s\n", line);
      continue;
    }
  }
  return 0;
}

static int process_files (int nfiles, char *file[])
{
  int i, j;
  struct stat st_h;
  FILE *out = NULL;
  char pname[LINELEN], fname[LINELEN];
  int plen, flen;

  report_inprogress (header_file);

  sprintf (pname, "%s/", proto_dir);
  sprintf (fname, "%s/", source_dir);
  plen = strlen (pname);
  flen = strlen (fname);

  if (stat (header_file, &st_h))
    ZERO_MTIME (st_h);

  /* Have to 'sort -u' them.  Break out the bubble sort. (: */
  for (i=0; i<nfiles; i++) {
    for (j=i+1; j<nfiles; j++) {
      char *tmp;
      int k = strcmp (file[i], file[j]);
      if (k == 0)		/* duplicate -- mark it */
	*file[i] = '\0';
      else if (k > 0) {
	tmp = file[i]; file[i] = file[j]; file[j] = tmp;
      }
    }
  }
  /* Remove duplicates marked above. */
  for (j=i=0; j<nfiles; i++,j++) {
    while (!*file[j] && j < nfiles)
      j++;
    if (i < j && j < nfiles)
      file[i] = file[j];
  }
  nfiles = i;

  for (i=0; i<nfiles; i++) {
    struct stat st_f, st_p;
    int need_out = 0;
    int l = strlen (file[i]);
    char *input = fname;

    if (file[i][l-2] == '.' && file[i][l-1] == 'o')
      file[i][l-1] = 'c';

    strcpy (pname+plen, file[i]);
    strcpy (fname+flen, file[i]);

    ZERO_MTIME (st_p);

    if (stat (input, &st_f)) {
      /* look again: source file may be in build directory */
      input = file[i];
      if (stat (input, &st_f)) {
	fprintf (stderr, "missing input file: %s\n", input);
	return -1;
      }
    }
    if (force || NEWER (st_f, st_h) || stat (pname, &st_p)) {
      need_out = 1;
      if (force || NEWER (st_f, st_p)) {
	/* check to see if proto is needed... */
	FILE *f1in, *f1out;
	f1in = xfopen (input, "r");
	f1out = xfopen (PROTO_TEMP, "w");
	process_one_file (f1in, f1out);
	fclose (f1in); fclose (f1out);
	/*
	 * If the .proto/foo/bar.c file was not touched,
	 * but the final include/foo_proto.h file does not
	 * exist, we still need to create the final output! 
	 * (Elrond)
	 */
	need_out = move_if_change (PROTO_TEMP, pname) || (st_h.st_mtime==0);
      }
    }
    if (!out && (force || need_out)) {
      out = open_header();
      for (j=0; j<i; j++)
	output_one_file (file[j], out);
    }

    if (out)
      output_one_file (file[i], out);
  }

  if (out)
    close_header (out);
  else
    report_finished (0);
  return 0;
}

int main (int argc, char *argv[])
{
  int c;
  extern char *optarg;
  extern int optind;

  progname = strrchr (argv[0], '/');
  if (!progname++)
    progname = argv[0];

  /* this will go away soon (: */
  if (argc>1 && strstr (argv[1], "awk")) {
    /*
      fprintf (stderr, "%s: legacy argument '%s' ignored\n", progname, argv[0]);
    */
    argv++; argc--;
  }

  while ((c = getopt (argc, argv, "h:s:f")) != -1) {
    switch (c) {
    case 'h':
      header_def = optarg; break;
    case 's':
      source_dir = optarg; break;
    case 'f':
      force = 1; break;
    default:
      usage();
    }
  }

  if (argc < 2)
    usage();

  header_file = argv[optind++];

  return process_files (argc-optind, argv+optind);
}
