/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Main SMB server routines
   Copyright (C) Andrew Tridgell 1992-1998,2001
   Copyright (C) Mark Cave-Ayland <mca198@ecs.soton.ac.uk>  2001,
   Copyright (C) Elrond <elrond@samba-tng.org>     2001.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "libsamba.h"
#include "services.h"
#include "pathnames.h"
#include "others.h"

fstring pipe_name;

static void load_services(char *pszFname);
static BOOL mdo_section(char *pszSectionName);
static BOOL mdo_parameter(char *pszParmName, char *pszParmVaue);
static BOOL start_services(void);

/* Declare an array for the service table */
SERVICE_TABLE srv_list[MAX_SERVICES];
int srv_index=-1;


/*************************************************************************
 initialise an msrpc service
 *************************************************************************/
static void service_init(char* service_name)
{
	add_msrpc_command_processor(pipe_name, "ntsvcs", api_svcctl_rpc,
				    NULL);

	generate_wellknown_sids();
}

/****************************************************************************
  reload the services file
  **************************************************************************/
static BOOL reload_msrpc(BOOL test)
{
	BOOL ret;

	if (lp_loaded()) {
		pstring fname;
		pstrcpy(fname,lp_configfile());
		if (file_exist (fname, NULL) && !strcsequal (fname, pathname_smb_conf)) {
			pathname_smb_conf = fname;
			test = False;
		}
	}

	reopen_logs();

	if (test && !lp_file_list_changed())
		return(True);

	lp_killunused(NULL);

	ret = lp_load (pathname_smb_conf, False, False, True);

	/* perhaps the config filename is now set */
	if (!test)
		reload_msrpc(True);

	reopen_logs();

	load_interfaces();

	return(ret);
}

/****************************************************************************
  main program
****************************************************************************/
static int main_init(int argc,char *argv[])
{
	fstring svcfile;
#ifdef HAVE_SET_AUTH_PARAMETERS
	set_auth_parameters(argc,argv);
#endif

#ifdef HAVE_SETLUID
	/* needed for SecureWare on SCO */
	setluid(0);
#endif

	TimeInit();

	setup_logging(argv[0],False);
	fstrcpy(pipe_name, "svcctl");
	slprintf(debugf, sizeof(debugf), "%s/log.%s", pathname_log_dir, pipe_name);
	slprintf (svcfile, sizeof(fstring)-1,
		  "%s/smb.services", pathname_conf_dir);

	load_services (svcfile);
	start_services();

	return 0;
}


static BOOL start_services(void)
{
	/* Start the services which have been marked as automatic
	in the service table */

	int i;

	for (i=0; i<=srv_index; i++) {

		/* If type set to Automatic... */
		if (srv_list[i].startup==0) {
			if (smbrun(srv_list[i].start, "/tmp/foo", False)==0) {
				srv_list[i].status=1;
				}	
			}
	}

	return True;
}


static void load_services(char *pszFname)
{
	/* Obtain the service list from smb.services and load into
	the array. When these services are started and stopped
	from Server Manager, the relevant script is executed. */

	pstring n2;
	BOOL bRetVal;

	/* Process the services file */
	pstrcpy(n2, pszFname);	
	bRetVal = pm_process(n2, mdo_section, mdo_parameter);
	
	return;
}


static BOOL mdo_section(char *pszSectionName)
{
	/* Load name into the array */
	srv_index++;
	fstrcpy(srv_list[srv_index].name, pszSectionName);

	/* printf("Section name: %s\n", pszSectionName); */

	return True;
}


static BOOL mdo_parameter(char *pszParmName, char *pszParmValue)
{
	/* Load parameters into the service table */
	if (strcmp(pszParmName, "name")==0) {
		fstrcpy(srv_list[srv_index].fullname, pszParmValue);
		}

	if (strcmp(pszParmName, "startup")==0) {

		/* Automatic or manual? */
		if (strcmp(pszParmValue, "auto")==0) {
			srv_list[srv_index].startup=0;
			} else {

			/* Default to manual */
			srv_list[srv_index].startup=1;
			}
	}

	if (strcmp(pszParmName, "pipe")==0) {
		fstrcpy(srv_list[srv_index].pipe, pszParmValue);
		}

	if (strcmp(pszParmName, "start script")==0) {
		fstrcpy(srv_list[srv_index].start, pszParmValue);
		}	

	if (strcmp(pszParmName, "stop script")==0) {
		fstrcpy(srv_list[srv_index].stop, pszParmValue);
		}

	/* printf("Param: %s   Value: %s\n", pszParmName, pszParmValue); */

	return True;
}


static msrpc_service_fns fn_table =
{
	NULL,
	service_init,
	reload_msrpc,
	main_init,
	NULL
};

msrpc_service_fns *get_service_fns(void)
{
	return &fn_table;
}
